
/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2000). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
/* Header file for image array allocations */

#define free_vector(x) free(x)

void exit_with_error();
void free_matrix(unsigned char **p,int m);
void free_matrix_int(int **p,int m);
void free_matrix_float(float **p,int m);
void free_matrix_complex(_COMPLEX **p,int m);
unsigned char *allocate_vector(int m,int size);
int *allocate_vector_int(int m,int size);
float  *allocate_vector_float(int m,int size);
_COMPLEX  *allocate_vector_complex(int m,int size);
int **allocate_matrix_int(int m,int n,int size);
float **allocate_matrix_float(int m,int n,int size);
unsigned char **allocate_matrix(int m,int n,int size);
_COMPLEX  **allocate_matrix_complex(int m,int n,int size);

void exit_with_error()
{
	exit(FALSE);
}


void free_matrix(unsigned char **p,int m)
{
	int i;

	for(i=0;i<m;++i)
        {
           // cout<<"Freeing "<<i<<"th row \n";
		free_vector(p[i]);
        }
	free_vector(p);
}

void free_matrix_int(int **p,int m)
{
	int i;

	for(i=0;i<m;++i)
		free_vector(p[i]);
	free_vector(p);
}

void free_matrix_float(float **p,int m)
{
	int i;

	for(i=0;i<m;++i)
		free_vector(p[i]);
	free_vector(p);
}

void free_matrix_complex(_COMPLEX **p,int m)
{
	int i;

	for(i=0;i<m;++i)
		free_vector(p[i]);
	free_vector(p);
}

unsigned char *allocate_vector(int m,int size)
{
	unsigned char *p;

	if ((p=(unsigned char *)calloc((unsigned) m,(unsigned) size))==NULL)
		cout<< "Insufficient memory\n";
	return(p);
}

int *allocate_vector_int(int m,int size)
{
	int *p;

	if ((p=(int *) calloc((unsigned) m,(unsigned) size))==NULL)
		cout<< "Insufficient memory\n";
	return(p);
}


float *allocate_vector_float(int m,int size)
{
	float *p;

	if ((p=(float *) calloc((unsigned) m,(unsigned) size))==NULL)
		cout<< "Insufficient memory\n";
	return(p);
}

_COMPLEX  *allocate_vector_complex(int m,int size)
{
	_COMPLEX *p;

	if ((p=(_COMPLEX  *) calloc((unsigned) m,(unsigned) size))==NULL)
		cout<< "Insufficient memory\n";
	return(p);
}

unsigned char **allocate_matrix(int m,int n,int size)
{
	unsigned char **p;
	int i;

	p=(unsigned char **) allocate_vector(m,sizeof(unsigned char *));
	for (i=0;(i==0 || (i>0 && p[i-1]!=NULL)) && i<m;++i)
		p[i]=allocate_vector(n,size);
	if (i!=m || p[m-1]==NULL) {
		free_matrix(p, m);
		p=NULL;
	}
	return(p);
}

int **allocate_matrix_int(int m,int n,int size)
{
	int **p;
	int i;

	p=(int **) allocate_vector_int(m,sizeof(int *));
	for (i=0;(i==0 || (i>0 && p[i-1]!=NULL)) && i<m;++i)
		p[i]=allocate_vector_int(n,size);
	if (i!=m || p[m-1]==NULL) {
		free_matrix_int(p, m);
		p=NULL;
	}
	return(p);
}

float **allocate_matrix_float(int m,int n,int size)
{
	float **p;
	int i;

	p=(float **) allocate_vector_float(m,sizeof(float *));
	for (i=0;(i==0 || (i>0 && p[i-1]!=NULL)) && i<m;++i)
		p[i]=allocate_vector_float(n,size);
	if (i!=m || p[m-1]==NULL) {
		free_matrix_float(p, m);
		p=NULL;
	}
	return(p);
}

_COMPLEX  **allocate_matrix_complex(int m,int n,int size)
{
	_COMPLEX  **p;
	int i;

	p=(_COMPLEX  **) allocate_vector_complex (m,sizeof(_COMPLEX  *));
	for (i=0;(i==0 || (i>0 && p[i-1]!=NULL)) && i<m;++i)
		p[i]=allocate_vector_complex (n,size);
	if (i!=m || p[m-1]==NULL) {
		free_matrix_complex(p, m);
		p=NULL;
	}
	return(p);
}

/*

int myopen(fname)
char *fname;
{
	int i, fd;
	char s[80];

	for (i=0; fname[i]!='\0' && fname[i]!='/'; ++i) ;
	if (fname[i]=='\0') {
		for (i=0; path[i]!=NULL && (i==0 || fd==EOF); ++i) {
			sprintf(s, "%s/%s", path[i], fname);
			fd=open(s, 0);
		}
	} else fd=open(fname, 0);
	return(fd);
}

FILE *myfopen(fname)
char *fname;
{
	int i;
	char s[80];
	FILE *fp, *fopen();

	for (i=0; fname[i]!='\0' && fname[i]!='/'; ++i) ;
	if (fname[i]=='\0') {
		for (i=0; path[i]!=NULL && (i==0 || fp==NULL); ++i) {
			sprintf(s, "%s/%s", path[i], fname);
			fp=fopen(s, "r");
		}
	} else fp=fopen(fname, "r");
	return(fp);
}
*/


