
/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2000). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
//signal.h describes 1D signal class

class _signal{
      float *rval;
      float *ival;
      int sg_no;
      int domain;
      float xmin; //span of domain
      float xmax;
     public:
            _signal();
            _signal(int num);
            _signal(float val,int num);
            _signal(float *val,int num,int dm);
            ~_signal(){}

       void  read();
       void  write();
       void  save(char *name){
                 ofstream fout(name);

                 if(!fout){
                   cout<<"Cannot create output file \n";
                   return;
                  }

                 int i;

                 fout<<sg_no<<"\n";
                 fout<<domain<<"\n";
                 for(i=0;i<sg_no;i++)
                 if(domain==FREQ) fout<<rval[i]<<" +j "<<ival[i]<<"\n";
                 else
                 fout<<i<<" "<<rval[i]<<"\n";
                 fout.close();
                 }

       void  read(char *name){
                 ifstream fin(name);

                 if(!fin){
                   cout<<"Cannot open input file \n";
                   return;
                  }

                 int i,j;
                 int tsg_no;
                 int dm;
                 char dum[8];

                 fin>>tsg_no;
                 fin>>dm;
                 step(0,tsg_no);
                 for(i=0;i<tsg_no;i++)
                 if(dm==FREQ) fin>>j>>rval[j]>>dum>>ival[j];
                 else
                 fin>>j>>rval[j];
                 fin.close();
                 }

       void free(){
                    if(rval!=NULL) {
                         delete rval;
                         rval=NULL;
                         }
                    if(ival!=NULL) 
                       {
                           delete ival;
                           ival=NULL;
                       }
                  }

       void  set(float *val,int num,int dm);//general loading
       void  set(float *real,float *imag,int num,int dm);//general loading
       void  step(float val,int num); //step function
       void  ramp(float a,float b,int num);//ramp function
       void  gauss(float mu,float sd,int num);//gaussian function
       void  gauss_der1(float mu,float sd,int num);//Ist der. of gaussian function
       void  gauss_der2(float mu,float sd,int num);//IInd der. of gaussian function
       void set_domain(int dm){if(dm==FREQ)cout<<"Domain FREQUENCY not settable \n";
                               else if((dm<0)||(dm>4)) domain=OTHER;
                               else domain=dm;
                              }

        int  get_num(){return(sg_no);}
        int  get_domain(){return(domain);}
        float *get_r(){return(rval);}
        float *get_i(){return(ival);}
        float get_xmin(){return(xmin);}
        float get_xmax(){return(xmax);}


       _signal operator+(_signal ob);
       _signal operator-(_signal ob);
       _signal operator*(_signal ob);//dot product
       _signal operator/(_signal ob);
       _signal operator%(_signal ob); //convolution
       _signal operator&&(_signal ob);//min op
       _signal operator||(_signal ob);//max op
       _signal &operator=(_signal ob);

       _signal operator+(float val);
       _signal operator-(float val);
       _signal operator*(float val);
       _signal operator/(float val);
       _signal operator%(float val);//Thrsholded by val
       _signal operator|(int pos); //truncate after pos
       _signal operator<(int pos); //shift left to  pos
       _signal operator>(int pos); //shift right to  pos

       float operator&(int pos);//extract the pos field

       float sum(){
       int i;
       float sum=0;

       for(i=0;i<sg_no;i++) sum+=rval[i];
       return(sum);
       }

       float sqr_sum()
       {
       int i;
       float sum=0;

       for(i=0;i<sg_no;i++) sum+=sqr(rval[i]);
       return(sum);
       }

       float avg()
       {
        return(sum()/(float)sg_no);
       }

       float sd()
       {
        return((float)sqrt((double)(sqr_sum()/(float)sg_no-sqr(avg()))));
       }

       float dm_sum()
       {
       int i;
       float sum=0;

       for(i=0;i<sg_no;i++) sum+=rval[i]*(float)i;
       return(sum);
       }

       float dm_sqr_sum()
       {
       int i;
       float sum=0;

       for(i=0;i<sg_no;i++) sum+=rval[i]*sqr((float)i);
       return(sum);
       }

       float dm_avg()
       {
        return(dm_sum()/sum());
       }

       float dm_sd()
       {
        return((float)sqrt((double)(dm_sqr_sum()/sum()-sqr(dm_avg()))));
        }

       _signal normalise()
       {
        _signal temp(sg_no);
         float area;
         int i;

         area=sum();
         for(i=0;i<sg_no;i++){
         temp.rval[i]=rval[i]/area;
         }
         return(temp);
       }

       _signal cumulative()
       {
        _signal temp(sg_no);
         int i;

         temp.rval[0]=rval[0];
         for(i=1;i<sg_no;i++){
         temp.rval[i]=temp.rval[i-1]+rval[i];
         }
         return(temp);
       }

       int get_thresh(float thresh=0.5) 
       {
        _signal temp1,temp2;
         int i;

         if(thresh>=1.0){
         cout<<"WARNING: thresholding above or equal to 100% in the range \n";
         return(sg_no-1);
         }
         if(thresh<=0){
         cout<<"WARNING: thresholding below 0% in the range \n";
         return(0);
         }
         temp1=normalise();
         temp2=temp1.cumulative();
         temp1.free();

         for(i=0;i<sg_no;i++)
         if(temp2.rval[i]>thresh) 
          break;
          
          temp2.free();
          return(i);
         }

      int getCount(float thresh) // Implemented on 10.4.2002
          {
            int count=0;
            int i;

         for(i=0;i<sg_no;i++)
         if(rval[i]>thresh) count++; 
          return count;
          }

     void   smooth();
    _signal valley();
    _signal gauss_smooth(float sd);
    _signal gauss_derivative(float sd);
    _signal ssf(float sd,int mode);
    _signal ssf_d(float sd,int mode);
    _signal zcross();
    _signal zcross_to_negative();
    _signal zcross_to_positive();
    _signal derivative();
    _signal dft();
    _signal idft();
    _signal convert_to_phasor();
    _signal convert_from_phasor();

        /*

       _signal integrate();
       _signal differentiate();
       _signal zcross();
      */
     };
     _signal::_signal()
      {
      rval=NULL;
      ival=NULL;
      domain=OTHER;
      xmin=xmax=0;
      sg_no=0;
      }

     _signal::_signal(int num)
     {
      int i;

      if(num==0){
      rval=NULL;
      ival=NULL;
      domain=OTHER;
      xmin=xmax=0;
      sg_no=0;
      }
      else{
       rval=new float[num];

      for(i=0;i<num;i++) rval[i]=0;
          ival=NULL;
          domain=OTHER;
          sg_no=num;
          xmin=0;
          xmax=num;
     }

      }

     _signal::_signal(float val,int num)
     {int i;
       rval=new float[num];

      for(i=0;i<num;i++) rval[i]=val;
          ival=NULL;
          domain=OTHER;
          sg_no=num;
          xmin=0;
          xmax=num;
     }

     _signal::_signal(float *val,int num,int dm)
     {
      if(dm==FREQ) {
         cout<<"Frequency domain not initialisable \n";
         rval=NULL;
         ival=NULL;
         domain=OTHER;
         xmin=xmax=0;
         sg_no=0;
         }
      else {
          rval=new float[num];
          int i;

          for(i=0;i<num;i++) rval[i]=val[i];
          ival=NULL;
          domain=dm;
          sg_no=num;
          xmin=0;
          xmax=num;
          }
      }         
       

       void  _signal::read()
         {
          int i;
    
          cout<<"No. of fields \n";
          cin>>sg_no;
          cout<<"Domain? --TIME(0),SPACE(1),OTHER(4)";
          cin>>domain;
          if(domain==3){
          cout<<"Domain Frequency not acceptable as input \n";
          sg_no=0;
          domain=0;
          return;
          }
          else if((domain<0)||(domain>4)) domain=OTHER;

          rval=new float[sg_no];

          for(i=0;i<sg_no;i++){
          cout<<i<<" the Field? \n";
          cin>>rval[i];
          }
          //return;
         }


       void  _signal::write()
         {
          int i;

          for(i=0;i<sg_no;i++){
          if(domain==FREQ) cout<<rval[i]<<" +j "<<ival[i]<<"\n";
          else    cout<<rval[i]<<"\n";
         }
         }

       void  _signal::set(float *val,int num,int dm=OTHER)//general loading
       {

      if(dm==FREQ) {
         cout<<"Frequency domain not initialisable \n";
         rval=NULL;
         ival=NULL;
         domain=OTHER;
         xmin=xmax=0;
         sg_no=0;
         }
      else {
          rval=new float[num];
          int i;

          for(i=0;i<num;i++) rval[i]=val[i];
          ival=NULL;
          domain=dm;
          sg_no=num;
          xmin=0;
          xmax=num;
          }
       }

 void  _signal::set(float *real,float *imag,int num,int dm)//general loading
 {
            rval=real;
            ival=imag;
            domain=dm;
            sg_no=num;
            xmin=0;
            xmax=num;
 }




       void  _signal::step(float val,int num) //step function
       {
        int i;

          rval=new float[num];
          for(i=0;i<num;i++) rval[i]=val;
          ival=NULL;
          sg_no=num;
          xmin=0;
          xmax=num;
       }

      void _signal::ramp(float a,float b,int num)
      {
        int i;

          rval=new float[num];
          for(i=0;i<num;i++) rval[i]=a+(float)i*(b-a)/(float)num;
          ival=NULL;
          sg_no=num;
          xmin=0;
          xmax=num;
      }

     void  _signal::gauss(float mu,float sd,int num)//gaussian function
     {
        int i;
        float x;

          rval=new float[num];
          for(i=0;i<num;i++) 
          {
          x=((float)i-mu)/sd;
          rval[i]= (float)exp(-sqr((double) x)/2.0)/(sd*(float)sqrt(2.0*M_PI));
          }
          ival=NULL;
          sg_no=num;
          xmin= 0;
          xmax= (float)num;
      }
            


     void  _signal::gauss_der1(float mu,float sd,int num)//gaussian function
     {
        int i;
        float x;

          rval=new float[num];
          for(i=0;i<num;i++) 
          {
          x=((float)i-mu)/sd;
          rval[i]= (float)exp(-sqr((double) x)/2.0)/(sd*(float)sqrt(2.0*M_PI));
          rval[i]=rval[i]*(-x/sqr(sd));
          }
          ival=NULL;
          sg_no=num;
          xmin= 0;
          xmax= (float)num;
      }
            

     void  _signal::gauss_der2(float mu,float sd,int num)//gaussian function
     {
        int i;
        float x;

          rval=new float[num];
          for(i=0;i<num;i++) 
          {
          x=((float)i-mu)/sd;
          rval[i]= (float)exp(-sqr((double) x)/2.0)/(sd*(float)sqrt(2.0*M_PI));
          rval[i]=rval[i]*(sqr(x/sd)-1)/sqr(sd);
          }
          ival=NULL;
          sg_no=num;
          xmin= 0;
          xmax= (float)num;
      }
            
       _signal _signal::operator+(_signal ob)
       {
        int i;
        _signal temp(0,max(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in addition \n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in addition \n";
        for(i=0;i<sg_no;i++) temp.rval[i]+=rval[i];
        for(i=0;i<ob.sg_no;i++) temp.rval[i]+=ob.rval[i];
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) temp.ival[i]=0;

        for(i=0;i<sg_no;i++) temp.ival[i]+=ival[i];
        for(i=0;i<ob.sg_no;i++) temp.ival[i]+=ob.ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator-(_signal ob)
       {
        int i;
        _signal temp(0,max(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in subtraction\n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in subtraction\n";
        for(i=0;i<sg_no;i++) temp.rval[i]-=rval[i];
        for(i=0;i<ob.sg_no;i++) temp.rval[i]-=ob.rval[i];
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) temp.ival[i]=0;
        for(i=0;i<sg_no;i++) temp.ival[i]-=ival[i];
        for(i=0;i<ob.sg_no;i++) temp.ival[i]-=ob.ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator*(_signal ob)
       {
        int i;
        _signal temp(0.0,min(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in multiplication\n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in multiplication\n";
        for(i=0;i<temp.sg_no;i++) temp.rval[i]=rval[i]*ob.rval[i];
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) {
           temp.rval[i]=rval[i]*ob.rval[i]-ival[i]*ob.ival[i];
           temp.ival[i]=rval[i]*ob.ival[i]+ival[i]*ob.rval[i];
        }
        }
        temp.domain=domain;
        return(temp);
       }

       _signal _signal::operator/(_signal ob)
       {
        int i;
        float mag;
        _signal temp(0.0,min(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in division\n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in division\n";
        for(i=0;i<temp.sg_no;i++) temp.rval[i]=rval[i]/ob.rval[i];
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) {
           mag=sqr(ob.rval[i])+sqr(ob.ival[i]);
           temp.rval[i]=(rval[i]*ob.rval[i]+ival[i]*ob.ival[i])/mag;
           temp.ival[i]=(ival[i]*ob.rval[i]-rval[i]*ob.ival[i])/mag;
        }
        }
        temp.domain=domain;
        return(temp);
       }

       _signal _signal::operator%(_signal ob) //convolution
       {
        int i,j;
        int mid_pos;
        float sum;
        _signal temp(0,sg_no);

         mid_pos=ob.sg_no/2;
         for(i=0;i<sg_no;i++){
         sum=0;
         for(j=0;j<ob.sg_no;j++){
         if(((i-mid_pos+j)>=0)&&((i-mid_pos+j)<sg_no))
         sum+=rval[i-mid_pos+j]*ob.rval[j];
         }
         temp.rval[i]=sum;
         }

        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator&&(_signal ob) //min operation
       {
        int i;
        _signal temp(0,min(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in min operation\n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in min operation\n";
        for(i=0;i<temp.sg_no;i++) temp.rval[i]=min(rval[i],ob.rval[i]);
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) temp.rval[i]=min(ival[i],ob.ival[i]);
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator||(_signal ob) //max operation
       {
        int i;
        _signal temp(0,max(sg_no,ob.sg_no));

        if(domain!=ob.domain) cout<<"WARNING:Domain mismatch in min operation\n";
        if(xmin!=ob.xmin) cout<<"WARNING:Lower range mismatch in min operation\n";
        for(i=0;i<sg_no;i++) temp.rval[i]=max(temp.rval[i],rval[i]);
        for(i=0;i<ob.sg_no;i++) temp.rval[i]=max(temp.rval[i],ob.rval[i]);
        if((domain==FREQ)&&(ob.domain==FREQ)){
        temp.ival=new float[temp.sg_no];

        for(i=0;i<temp.sg_no;i++) temp.ival[i]=0;
        for(i=0;i<sg_no;i++) temp.ival[i]=max(temp.ival[i],ival[i]);
        for(i=0;i<ob.sg_no;i++) temp.ival[i]=max(temp.ival[i],ob.ival[i]);
        }
        temp.domain=domain;
        return(temp);
       }

       _signal &_signal::operator=(_signal ob) // assignment operation
       {
        int i;

        if(rval!=NULL) delete rval;
        rval=new float[ob.sg_no];
        for(i=0;i<ob.sg_no;i++) rval[i]=ob.rval[i];
        if(ob.domain==FREQ){
        if(ival!=NULL) delete ival;
        ival=new float[ob.sg_no];
        for(i=0;i<ob.sg_no;i++) ival[i]=ob.ival[i];
        }
        xmin=ob.xmin;
        xmax=ob.xmax;
        sg_no=ob.sg_no;
        domain=ob.domain;
        return(*this);
       }



       _signal _signal::operator+(float val)
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) temp.rval[i]=rval[i]+val;
        if(domain==FREQ){
        temp.ival=new float[temp.sg_no];
        for(i=0;i<sg_no;i++)temp.ival[i]=ival[i];
        }
        temp.domain=domain;
        return(temp);
       }

       _signal _signal::operator-(float val)
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) temp.rval[i]=rval[i]-val;
        if(domain==FREQ){
        temp.ival=new float[temp.sg_no];
        for(i=0;i<sg_no;i++)temp.ival[i]=ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator*(float val)
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) temp.rval[i]=rval[i]*val;
        if(domain==FREQ){
        temp.ival=new float[temp.sg_no];
        for(i=0;i<sg_no;i++)temp.ival[i]=ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator/(float val)
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) temp.rval[i]=rval[i]/val;
        if(domain==FREQ){
        temp.ival=new float[temp.sg_no];
        for(i=0;i<sg_no;i++)temp.ival[i]=ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator%(float val) //Thresholded by val
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) temp.rval[i]=(rval[i]>val)?rval[i]:0;
        if(domain==FREQ){
        temp.ival=new float[temp.sg_no];
        for(i=0;i<sg_no;i++)temp.ival[i]=(temp.rval[i]>0)?ival[i]:0;
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator|(int pos) //truncate after pos
       {
        int i;
        _signal temp(0,pos);

        for(i=0;i<pos;i++) temp.rval[i]=rval[i];
        if(domain==FREQ){
        temp.ival=new float[pos];
        for(i=0;i<pos;i++)temp.ival[i]=ival[i];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator<(int pos) //left shift by pos
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) {
        if((i+pos)<sg_no)
        temp.rval[i]=rval[i+pos];
        }
        if(domain==FREQ){
        temp.ival=new float[sg_no];
        if((i+pos)<sg_no)
        temp.ival[i]=ival[i+pos];
        }
        temp.domain=domain;
        return(temp);
       }


       _signal _signal::operator>(int pos) //right shift by pos
       {
        int i;
        _signal temp(0,sg_no);

        for(i=0;i<sg_no;i++) {
        if((i-pos)>=0)
        temp.rval[i]=rval[i-pos];
        }
        if(domain==FREQ){
        temp.ival=new float[sg_no];
        if((i-pos)>=0)
        temp.ival[i]=ival[i-pos];
        }
        temp.domain=domain;
        return(temp);
       }

       float  _signal::operator&(int pos) //extract at  pos
       {
        if(pos<sg_no) return(rval[pos]);
        else {
        cout<<"No element at"<<pos<<"\n";
        return(0);
        }
       }


		void _signal::smooth()
		{int i;

                if(domain!=FREQ)
		for(i=2;i<sg_no;i++)
		rval[i]= (rval[i-1]+rval[i]+rval[i+1])/3;
                else
		for(i=2;i<sg_no;i++)
                {
		ival[i]= (ival[i-1]+ival[i]+ival[i+1])/3;
		ival[i]= (ival[i-1]+ival[i]+ival[i+1])/3;
                }
                
		}


  


    _signal _signal::valley()
            {
             _signal temp;


                 int i,j,k;
                 float temppit[256];

                 
                  k=0;
	      for(i=1;i<sg_no-1;i++) 
               if((rval[i]<rval[i-1])&&(rval[i]<=rval[i+1]))
                       temppit[k++]=(float)i;

                  if(k!=0){
                  float *pit;

                  pit= new float[k];
                  for(i=0;i<k;i++) pit[i]=temppit[i];
                  temp.set(pit,k,OTHER);
                  }
                  return(temp);
                    }


    _signal _signal::zcross_to_negative()// +ve to -ve transitions
     {
                 _signal temp;
                 int i,j,k;
                 float tcross[256];

              k=0;
	      for(i=1;i<sg_no-1;i++) 
               if((rval[i-1]>=0)&&(rval[i]<0))
                       tcross[k++]=(float)i;

                 float *cross;
                 cross= new float[k];

                 for(i=0;i<k;i++) cross[i]=tcross[i];

                  temp.set(cross,k,OTHER);
                  return(temp);
     }

    _signal _signal::zcross_to_positive()// -ve to +ve transitions
     {
                 _signal temp;
                 int i,j,k;
                 float tcross[256];

              k=0;
	      for(i=1;i<sg_no-1;i++) 
               if((rval[i-1]<0)&&(rval[i]>=0))
                       tcross[k++]=(float)i;

                 float *cross;
                 cross= new float[k];

                 for(i=0;i<k;i++) cross[i]=tcross[i];

                  temp.set(cross,k,OTHER);
                  return(temp);
     }

    _signal _signal::zcross()// any transition
     {
                 _signal temp;
                 int i,j,k;
                 float tcross[256];

              k=0;
	      for(i=1;i<sg_no-1;i++) 
               if((rval[i-1]<0)&&(rval[i]>=0)|| (rval[i-1]>=0)&&(rval[i]<0))
                       tcross[k++]=(float)i;

                 float *cross;
                 cross= new float[k];

                 for(i=0;i<k;i++) cross[i]=tcross[i];

                  temp.set(cross,k,OTHER);
                  return(temp);
     }

     _signal _signal::gauss_smooth(float sd)
     {
       _signal temp,temp_g;

        if(sd==0) return(*this);
        temp_g.gauss(2*(int)sd,sd,4*(int)sd);
        temp=(*this)%temp_g;
        temp_g.free();
        return(temp);
     }

     _signal _signal::ssf(float sd,int mode=0)
     {
       _signal temp,temp_g,temp_z;
        int num;
        

        if(sd==0) return(temp);
        num=3*(int)sd+1;
        temp_g.gauss_der1(num/2,sd,num);
        temp=(*this)%temp_g;
         if(mode>0) temp_z=temp.zcross_to_positive();
         else if(mode<0) temp_z=temp.zcross_to_negative();
          else temp_z=temp.zcross();


        temp_g.free();
        temp.free();
        return(temp_z);
     }

     _signal _signal::ssf_d(float sd,int mode=0)
     {
       _signal temp,temp_d,temp_z;

        if(sd==0) return(temp);
        temp=gauss_smooth(sd);
        temp_d=temp.derivative();
         if(mode>0) temp_z=temp_d.zcross_to_positive();
         else if(mode<0) temp_z=temp_d.zcross_to_negative();
          else temp_z=temp_d.zcross();
        temp_d.free();
        temp.free();
        return(temp_z);
     }

    _signal _signal::gauss_derivative(float sd)
     {
       _signal temp,temp_g;

        if(sd==0) return(*this);
        temp_g.gauss_der1(2*(int)sd,sd,4*(int)sd);
        temp=(*this)%temp_g;
        temp_g.free();
        return(temp);
     }

    _signal _signal::derivative()
     {
       _signal temp(sg_no);
        int i;

        for(i=1;i<sg_no-1;i++)
           temp.rval[i]=(rval[i+1]-rval[i-1])/2.0;

        return(temp);
     }


    _signal _signal::dft()
      {
            _signal temp;
             int i,m,num=1;

            if(domain==FREQ){
            cout<<"dft of FREQ domain not allowed \n";
            return(temp);
            }
            
            m=(int)(log((double)sg_no)/log(2));
           num=num<<m;//num= pow(2,m);
           if(num<sg_no) {num=num<<1;
                          m++;}
            cout<<m<<" point FFT with  "<<num<<" data\n";

            temp.rval=new float[num];
            temp.ival=new float[num];
            temp.sg_no=num;
            temp.domain=FREQ;
            temp.xmin=0;
            temp.xmax=num;
         
             _COMPLEX *data;
              data=new _COMPLEX[num];
              for(i=0;i<sg_no;i++) data[i].set(rval[i],0);
              fft(data,m);
              for(i=0;i<num;i++) {
              temp.rval[i]=data[i].get_r();
              temp.ival[i]=data[i].get_i();
              }
              delete data;
              return(temp);
           }


    _signal _signal::idft()
      {
            _signal temp;
             int i,m,num=1;

            if(domain!=FREQ){
            cout<<"idft only applied to FREQ  domain  \n";
            return(temp);
            }
            
            m=(int)(log((double)sg_no)/log(2));
           num=num<<m;//num= pow(2,m);
           if(num<sg_no) m++;
             else if(num>sg_no) m--;

            cout<<m<<" point IFFT with  "<<num<<" data\n";

            temp.rval=new float[sg_no];
            temp.ival=new float[sg_no];
            temp.sg_no=num;
            temp.domain=OTHER;
            temp.xmin=0;
            temp.xmax=num;
         
             _COMPLEX *data;
              data=new _COMPLEX[num];
              for(i=0;i<sg_no;i++) data[i].set(rval[i],ival[i]);
              ifft(data,m);
              for(i=0;i<num;i++) {
              temp.rval[i]=data[i].get_r();
              temp.ival[i]=data[i].get_i();
              }
              delete data;
              return(temp);
           }

    _signal _signal::convert_to_phasor()
     {
            _signal temp;
             int i,m;

            if(domain!=FREQ){
            cout<<"phasor conversion is  only applied to FREQ  domain  \n";
            return(temp);
            }

            temp.rval=new float[sg_no];
            temp.ival=new float[sg_no];
            temp.sg_no=sg_no;
            temp.domain=FREQ;
            temp.xmin=0;
            temp.xmax=sg_no;
         
             _COMPLEX data,phasor;
              for(i=0;i<sg_no;i++) {
               data.set(rval[i],ival[i]);
               phasor=data.get_phasor();
               temp.rval[i]=phasor.get_r();
               temp.ival[i]=phasor.get_i();
               }
            
              return(temp);
     }


    _signal _signal::convert_from_phasor()
     {
            _signal temp;
             int i,m,num=1;

            if(domain!=FREQ){
            cout<<"conversion from phasor is  only applied to FREQ  domain  \n";
            return(temp);
            }

            temp.rval=new float[sg_no];
            temp.ival=new float[sg_no];
            temp.sg_no=num;
            temp.domain=FREQ;
            temp.xmin=0;
            temp.xmax=num;
         
             _COMPLEX data,phasor;
              for(i=0;i<sg_no;i++) {
               phasor.set(rval[i],ival[i]);
               data=phasor.convert_phasor();
               temp.rval[i]=data.get_r();
               temp.ival[i]=data.get_i();
               }
            
              return(temp);
     }
