
/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2000). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
// Class for extracting fields and frames from NTSC video stream.
// The data is in YCrCb (4:2:2) format.
class _NTSCVIDEO{
        int width,height;
	long int flength;
	double scale;
	//int nframes;
	//int nfields;
	int total_frame_no;
	int curr_frame_no;
	int interlace_mode; // _INTERLACE (1) or _DEINTERLACE (0)
	ifstream fin;

        void YCrCbtoRGB(double Y, double Cr, double Cb,
			double& R, double& G, double& B, double scale);
        _COLOR   readFrameFromFile(int long fposition);
         _COLOR  readFieldFromFile(int long fposition,int field_type);
        public:

               _NTSCVIDEO(){
		 width=720;
		 height=486;
		 scale=1.0;
		 interlace_mode=_DEINTERLACE;
		 curr_frame_no=0;
		 total_frame_no=0;
		 }
	       ~_NTSCVIDEO(){}
               float setScale(float val)
		     {
		      scale=val;
		     }
               int setInterlaceMode(int mode)
		   {
		     interlace_mode=mode;
		     return(interlace_mode);
		   }
               int getInterlaceMode()
		   {
		     return(interlace_mode);
		   }
               int getCurrentFrameNo()
		  {
		   return curr_frame_no;
		  }
                float getScale()
		   {
		    return scale;
		   }
		   float get_w()
		   {
		    return width;
		   }
		   float get_h()
		   {
		    return height;
		   }
		   
               _COLOR getCurrentFrame();
               _COLOR getPreviousFrame();
               _COLOR getNextFrame();
    int open(char *fname);
    int close();
    _COLOR getFrameNo(int num);
    _COLOR getField(int frame_num,int field_type);
	       };
 
// *********************************************************
int   _NTSCVIDEO::open(char *fname)
		 {
		   fin.open(fname);
		   if(!fin){
		            cout<<"Unable to open"<<fname<<"\n";
				   return(0);
		            }
			    //fin.seekg(ios::end);
			    //flength=fin.tellg();
                   //total_frame_no= (int)(flength/(2L * (long int)width * (long int) height));
		   total_frame_no=30; // WARNING!! Should be changed?
		   cout<<"Total No. of frames = "<<total_frame_no<<"\n";
		   return(1);
		 }
// *********************************************************
int   _NTSCVIDEO:: close()
		 {
		  if(fin) fin.close();
		 }
// *********************************************************

void _NTSCVIDEO::YCrCbtoRGB(double Y, double Cr, double Cb,
	                   double& R, double& G, double& B, double scale)
{
	// Convert
	R = scale*(Y + 1.3707*Cr);
	G = scale*(Y - 0.3365*Cb - 0.6982*Cr);
	B = scale*(Y + 1.7324*Cb);

	// Handle range overflow
	R = (R < 0.0 ? 0.0 : R);
	R = (R > 255.0 ? 255.0 : R);
	G = (G < 0.0 ? 0.0 : G);
	G = (G > 255.0 ? 255.0 : G);
	B = (B < 0.0 ? 0.0 : B);
	B = (B > 255.0 ? 255.0 : B);
}

// *********************************************************
 _COLOR   _NTSCVIDEO::readFrameFromFile(int long fposition)
      {
      _COLOR temp;
      int i,j,ind;
	double Cr, Cb, Y1, Y2;		// Input data values
	double R,G,B;				// Converted RGB values
	unsigned char bCr,bCb,bY1,bY2;
	unsigned char **red,**green,**blue;

//Position the file pointer 
      fin.seekg(fposition,ios::beg);

      cout<<"File positioned at "<<fposition<<"\n";


      temp.set(width,height);
      red=temp.get_A().get_pic();
      green=temp.get_B().get_pic();
      blue=temp.get_C().get_pic();

      cout<<"temporary _COLOR object created and pointers initialised \n"<<flush;

	for (i=0; i<height; i++) {

		// Split into two fields, if necessary
		if (interlace_mode==_INTERLACE)
			ind = ( (i & 0x01) ? (i/2): (height/2+i/2) );
		else
			ind = i;

		// Copy data
		for (j=0; j<width; j+=2)
		{
			// Read data
			//fin>>bCb>>bY1>>bCr>>bY2;
			fin.get(bCb);
			fin.get(bY1);
			fin.get(bCr);
			fin.get(bY2);
			      

			Cb=(double)bCb-128.0;
			Cr=(double)bCr-128.0;
			Y1=(double)bY1-16.0;
			Y2=(double)bY2-16.0;

			// First pixel
			YCrCbtoRGB(Y1,Cr,Cb,R,G,B,scale);
			red[ind][j]=(unsigned char) R;
			green[ind][j]=(unsigned char) G;
			blue[ind][j]=(unsigned char) B;

			// Second pixel
			YCrCbtoRGB(Y2,Cr,Cb,R,G,B,scale);
			red[ind][j+1]=(unsigned char) R;
			green[ind][j+1]=(unsigned char) G;
			blue[ind][j+1]=(unsigned char) B;
		}
         	}
		return temp;
     }

// *********************************************************
 _COLOR   _NTSCVIDEO::readFieldFromFile(int long fposition,int field_type)
      {
      _COLOR temp;
      int i,j,ind;
	double Cr, Cb, Y1, Y2;		// Input data values
	double R,G,B;				// Converted RGB values
	unsigned char bCr,bCb,bY1,bY2;
	unsigned char **red,**green,**blue;

//Position the file pointer 
      fin.seekg(fposition,ios::beg);

      cout<<"File positioned at "<<fposition<<"\n";


      temp.set(width,height/2);
      red=temp.get_A().get_pic();
      green=temp.get_B().get_pic();
      blue=temp.get_C().get_pic();

      cout<<"temporary _COLOR object created and pointers initialised \n"<<flush;

	for (i=0; i<height; i++) {

		// Copy data
		for (j=0; j<width; j+=2)
		{
			// Read data
			//fin>>bCb>>bY1>>bCr>>bY2;
			fin.get(bCb);
			fin.get(bY1);
			fin.get(bCr);
			fin.get(bY2);
			      

                        if((i &0x01)==field_type)
			{
			 ind=i/2;
			Cb=(double)bCb-128.0;
			Cr=(double)bCr-128.0;
			Y1=(double)bY1-16.0;
			Y2=(double)bY2-16.0;

			// First pixel
			YCrCbtoRGB(Y1,Cr,Cb,R,G,B,scale);
			red[ind][j]=(unsigned char) R;
			green[ind][j]=(unsigned char) G;
			blue[ind][j]=(unsigned char) B;

			// Second pixel
			YCrCbtoRGB(Y2,Cr,Cb,R,G,B,scale);
			red[ind][j+1]=(unsigned char) R;
			green[ind][j+1]=(unsigned char) G;
			blue[ind][j+1]=(unsigned char) B;
			}
		}
         	}
		return temp;
     }
// *********************************************************
    _COLOR _NTSCVIDEO::getFrameNo(int num)
    {
     _COLOR temp;
     int long fposition;

      if((num<0) || (num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)num*(int long)width*(int long)height*2L); 
      temp=readFrameFromFile(fposition);
      curr_frame_no=num;

      return temp;
    }

// *********************************************************
    _COLOR _NTSCVIDEO::getCurrentFrame()
    {
     _COLOR temp;
     int long fposition;

      fposition=(int long) ((int long)curr_frame_no*(int long)width*(int long)height*2L); 
      temp=readFrameFromFile(fposition);

      return temp;
    }
// *********************************************************
    _COLOR _NTSCVIDEO::getNextFrame()
    {
     _COLOR temp;
     int long fposition;

      if(curr_frame_no==total_frame_no-1)
	 {
	   cout<<"At the END of the stream; returning NULL image!!! \n";
	   return temp;
	 }
	 else
	  curr_frame_no++;
      fposition=(int long) ((int long)curr_frame_no*(int long)width*(int long)height*2L); 
      temp=readFrameFromFile(fposition);

      return temp;
    }
// *********************************************************
    _COLOR _NTSCVIDEO::getPreviousFrame()
    {
     _COLOR temp;
     int long fposition;

      if(curr_frame_no==0)
	 {
	   cout<<"At the BEGINNING of the stream; returning NULL image!!! \n";
	   return temp;
	 }
	 else
	  curr_frame_no++;
      fposition=(int long) ((int long)curr_frame_no*(int long)width*(int long)height*2L); 
      temp=readFrameFromFile(fposition);

      return temp;
    }

// *********************************************************

    _COLOR _NTSCVIDEO::getField(int frame_num,int field_type)
    {
     _COLOR temp;
     int long fposition;

      if((frame_num<0) || (frame_num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)frame_num*(int long)width*(int long)height*2L); 
      temp=readFieldFromFile(fposition,field_type);
      curr_frame_no=frame_num;

      return temp;
    }
