/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2000). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
// Class for extracting fields and frames from QCIF video stream.
// The data is in YCrCb (4:2:0) format.
//First Y (176 x 144), U (88 x 72) and V (88 x 72)
class _QCIFVIDEO{
        int width,height;
	long int flength;
	double scale;
	//int nframes;
	int total_frame_no;
	int curr_frame_no;
	int long frame_size;
	ifstream fin;
	_GREY imY,imU,imV;
	unsigned char **picY,**picU,**picV;

        void YCrCbtoRGB(double Y, double Cr, double Cb,
			double& R, double& G, double& B, double scale);
         int     readYUVfromFile(int long fposition);
        _COLOR   readFrameFromFile(int long fposition);
         _GREY   readGreyFrameFromFile(int long fposition);
          _GREY  readUFrameFromFile(int long fposition);
          _GREY  readVFrameFromFile(int long fposition);
        public:

               _QCIFVIDEO(){
		 width=0;
		 height=0;
		 scale=1.0;
		 curr_frame_no=0;
		 total_frame_no=0;
		 picY=picU=picV=NULL;
		 frame_size=0L;
		 }
	       ~_QCIFVIDEO(){}
               float setScale(float val)
		     {
		      scale=val;
		     }
               int getCurrentFrameNo()
		  {
		   return curr_frame_no;
		  }
                float getScale()
		   {
		    return scale;
		   }
		   float get_w()
		   {
		    return width;
		   }
		   float get_h()
		   {
		    return height;
		   }
                  
                  int getTotalFrameNo()
                   {
                      return total_frame_no;
                   }
		   
               _COLOR getCurrentFrame();
               _COLOR getPreviousFrame();
               _COLOR getNextFrame();
    int open(char *fname,int w=176,int h=144);
    int close();
    _COLOR getFrameNo(int num);
    _GREY getGreyFrameNo(int num);
    _GREY getUFrameNo(int num);
    _GREY getVFrameNo(int num);
	       };
 
// *********************************************************
int   _QCIFVIDEO::open(char *fname,int w,int h)
		 {
		   int long bposition,eposition;
		   fin.open(fname);
		   if(!fin){
		            cout<<"Unable to open"<<fname<<"\n";
				   return(0);
		            }
			    bposition=fin.tellg();
			    fin.seekg(0,ios::end);
			    eposition=fin.tellg();
			    flength=eposition-bposition;
			    fin.seekg(0,ios::beg);
			    cout<<"Size of the file "<<flength<<"\n";
		   width=w;
		   height=h;
		   imY.set(w,h); imU.set(w/2,h/2); imV.set(w/2,h/2);
		   picY=imY.get_pic();
		   picU=imU.get_pic();
		   picV=imV.get_pic();
		   if((picY==NULL)||(picU==NULL)||(picV==NULL))
		    {
		     cout<<"Memory not allocated for the video stream !! \n";
		     return 0;
		    }
		    frame_size=(int long)w*(int long)h+(int long)w/2L*(int long)h/2L+(int long)w/2L*(int long)h/2L;
		   total_frame_no=flength/frame_size; // WARNING!! Should be changed?
		   cout<<" Total No. of frames = "<<total_frame_no<<"\n";
		   return(1);
		 }
// *********************************************************
int   _QCIFVIDEO:: close()
		 {
		  if(fin) fin.close();
		  imY.free();
		  imU.free();
		  imV.free();
		 }
// *********************************************************

void _QCIFVIDEO::YCrCbtoRGB(double Y, double Cr, double Cb,
	                   double& R, double& G, double& B, double scale)
{
	// Convert
	R = scale*(Y + 1.3707*Cr);
	G = scale*(Y - 0.3365*Cb - 0.6982*Cr);
	B = scale*(Y + 1.7324*Cb);

	// Handle range overflow
	R = (R < 0.0 ? 0.0 : R);
	R = (R > 255.0 ? 255.0 : R);
	G = (G < 0.0 ? 0.0 : G);
	G = (G > 255.0 ? 255.0 : G);
	B = (B < 0.0 ? 0.0 : B);
	B = (B > 255.0 ? 255.0 : B);
}
// *********************************************************

int _QCIFVIDEO::readYUVfromFile(int long fposition)
{
    int i,j;

      //Position the file pointer 
      fin.seekg(fposition,ios::beg);
      // cout<<"File positioned at "<<fposition<<"\n";

        for (i=0;i<height;++i){
		fin.read((char *)picY[i],width);
		if (fin.gcount()!=width){
	         	cout<<"reading error in getting a line for Y\n";
	                return(0);
			 }
			 }
            // cout<<"picY read \n"<<flush;
        for (i=0;i<height/2;++i){
		fin.read((char *)picU[i],width/2);
		if (fin.gcount()!=width/2){
	         	cout<<"reading error in getting a line for U\n";
	                return(0);
			 }
			 }
            // cout<<"picU read \n"<<flush;
        for (i=0;i<height/2;++i){
		fin.read((char *)picV[i],width/2);
		if (fin.gcount()!=width/2){
	         	cout<<"reading error in getting a line for V\n";
	                return(0);
			 }
			 }
            // cout<<"picV read \n"<<flush;
    return 1;
}

// *********************************************************
 _COLOR   _QCIFVIDEO::readFrameFromFile(int long fposition)
      {
      _COLOR temp;
      int i,j;
	double Cr, Cb, Y1, Y2,Y3,Y4;		// Input data values
	double R,G,B;				// Converted RGB values
	unsigned char **red,**green,**blue;


       readYUVfromFile(fposition);

      temp.set(width,height);
      red=temp.get_A().get_pic();
      green=temp.get_B().get_pic();
      blue=temp.get_C().get_pic();

      // cout<<"temporary _COLOR object created and pointers initialised \n"<<flush;

	for (i=0; i<height/2; i++) 
	{
        for (j=0; j<width/2; j++)
		{

			Y1=(double)picY[2*i][2*j]-16.0;
			Y2=(double)picY[2*i][2*j+1]-16.0;
			Y3=(double)picY[2*i+1][2*j]-16.0;
			Y4=(double)picY[2*i+1][2*j+1]-16.0;
			Cb=(double)picU[i][j]-128.0;
			Cr=(double)picV[i][j]-128.0;

			// First pixel
			YCrCbtoRGB(Y1,Cr,Cb,R,G,B,scale);
			red[2*i][2*j]=(unsigned char) R;
			green[2*i][2*j]=(unsigned char) G;
			blue[2*i][2*j]=(unsigned char) B;

			// Second pixel
			YCrCbtoRGB(Y2,Cr,Cb,R,G,B,scale);
			red[2*i][2*j+1]=(unsigned char) R;
			green[2*i][2*j+1]=(unsigned char) G;
			blue[2*i][2*j+1]=(unsigned char) B;

			// Third pixel
			YCrCbtoRGB(Y3,Cr,Cb,R,G,B,scale);
			red[2*i+1][2*j]=(unsigned char) R;
			green[2*i+1][2*j]=(unsigned char) G;
			blue[2*i+1][2*j]=(unsigned char) B;

			// Fourth pixel
			YCrCbtoRGB(Y4,Cr,Cb,R,G,B,scale);
			red[2*i+1][2*j+1]=(unsigned char) R;
			green[2*i+1][2*j+1]=(unsigned char) G;
			blue[2*i+1][2*j+1]=(unsigned char) B;

		}
         	}
		return temp;
     }


// *********************************************************
 _GREY   _QCIFVIDEO::readGreyFrameFromFile(int long fposition)
      {
      _GREY temp;
      int i,j;
	unsigned char **lum;


       readYUVfromFile(fposition);

      temp.set(width,height);
      lum=temp.get_pic();

      // cout<<"temporary _GREY object created and pointers initialised \n"<<flush;

	for (i=0; i<height; i++) 
        for (j=0; j<width; j++)
           lum[i][j]=picY[i][j];

		return temp;
     }

// *********************************************************
 _GREY   _QCIFVIDEO::readUFrameFromFile(int long fposition)
      {
      _GREY temp;
      int i,j;
	unsigned char **lum;


       readYUVfromFile(fposition);

      temp.set(width/2,height/2);
      lum=temp.get_pic();

      // cout<<"temporary _GREY object created for U Frame and pointers initialised \n"<<flush;

	for (i=0; i<height/2; i++) 
        for (j=0; j<width/2; j++)
           lum[i][j]=picU[i][j];

		return temp;
     }

// *********************************************************
 _GREY   _QCIFVIDEO::readVFrameFromFile(int long fposition)
      {
      _GREY temp;
      int i,j;
	unsigned char **lum;


       readYUVfromFile(fposition);

      temp.set(width/2,height/2);
      lum=temp.get_pic();

      // cout<<"temporary _GREY object created for V Frame and pointers initialised \n"<<flush;

	for (i=0; i<height/2; i++) 
        for (j=0; j<width/2; j++)
           lum[i][j]=picV[i][j];

		return temp;
     }

// *********************************************************
    _COLOR _QCIFVIDEO::getFrameNo(int num)
    {
     _COLOR temp;
     int long fposition;

      if((num<0) || (num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)num*frame_size); 
      temp=readFrameFromFile(fposition);
      curr_frame_no=num;

      return temp;
    }

// *********************************************************
    _GREY _QCIFVIDEO::getGreyFrameNo(int num)
    {
     _GREY temp;
     int long fposition;

      if((num<0) || (num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)num*frame_size); 
      temp=readGreyFrameFromFile(fposition);
      curr_frame_no=num;

      return temp;
    }

// *********************************************************
    _GREY _QCIFVIDEO::getUFrameNo(int num)
    {
     _GREY temp;
     int long fposition;

      if((num<0) || (num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)num*frame_size); 
      temp=readUFrameFromFile(fposition);
      curr_frame_no=num;

      return temp;
    }

// *********************************************************
    _GREY _QCIFVIDEO::getVFrameNo(int num)
    {
     _GREY temp;
     int long fposition;

      if((num<0) || (num>total_frame_no))
	 {
	   cout<<"Number improper NULL image returned!! \n";
	   return temp;
	 }
      fposition=(int long) ((int long)num*frame_size); 
      temp=readVFrameFromFile(fposition);
      curr_frame_no=num;

      return temp;
    }

// *********************************************************
    _COLOR _QCIFVIDEO::getCurrentFrame()
    {
     _COLOR temp;
     int long fposition;

      fposition=(int long) ((int long)curr_frame_no*frame_size); 
      temp=readFrameFromFile(fposition);

      return temp;
    }
// *********************************************************
    _COLOR _QCIFVIDEO::getNextFrame()
    {
     _COLOR temp;
     int long fposition;

      if(curr_frame_no==total_frame_no-1)
	 {
	   cout<<"At the END of the stream; returning NULL image!!! \n";
	   return temp;
	 }
	 else
	  curr_frame_no++;
      fposition=(int long) ((int long)curr_frame_no*frame_size); 
      temp=readFrameFromFile(fposition);

      return temp;
    }
// *********************************************************
    _COLOR _QCIFVIDEO::getPreviousFrame()
    {
     _COLOR temp;
     int long fposition;

      if(curr_frame_no==0)
	 {
	   cout<<"At the BEGINNING of the stream; returning NULL image!!! \n";
	   return temp;
	 }
	 else
	  curr_frame_no++;
      fposition=(int long) ((int long)curr_frame_no*frame_size); 
      temp=readFrameFromFile(fposition);

      return temp;
    }

