/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2001). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
class _COLVIDFRCODE{
  int width,height;
  int HRNG_SIZE,RNG_SIZE,L_GAP;
  float ERR_LB,ERR_TH;
  int aNbitQ,bNbitQ;
  int aNbitR,bNbitR;
  _QDFSTDECODE *frcYQ,*frcUQ,*frcVQ; // Pointer to array of I(Q) frames
  _RELFRCODE   *frcYR,*frcUR,*frcVR; // Pointer to array of Relative Frames
  int codeflag; // 1 or 0
  int videoflag; // 1 or 0
  int allocateflag; // 1 or 0
  _QCIFVIDEO vid;
   int startFrame;
   int endFrame;
   int totalVideoFrameNo;
   int totalCompressionFrameNo;
   int GOP;
   int totalQFrameNo;
   int totalRFrameNo;

       void allocateFractalCodes()
       {
           if(GOP<1) {
             cout<<"GOP NOT DEFINED \n";
             return;
           }
           frcYQ=(_QDFSTDECODE *)calloc(totalQFrameNo,sizeof(_QDFSTDECODE));
           frcUQ=(_QDFSTDECODE *)calloc(totalQFrameNo,sizeof(_QDFSTDECODE));
           frcVQ=(_QDFSTDECODE *)calloc(totalQFrameNo,sizeof(_QDFSTDECODE));
           if((frcYQ==NULL) || (frcUQ==NULL) || (frcVQ==NULL))
              {
                cout<<"Memory for Q codes NOT allocated \n";
                exit(0);
              }
           frcYR=(_RELFRCODE *)calloc(totalRFrameNo,sizeof(_RELFRCODE));
           frcUR=(_RELFRCODE *)calloc(totalRFrameNo,sizeof(_RELFRCODE));
           frcVR=(_RELFRCODE *)calloc(totalRFrameNo,sizeof(_RELFRCODE));
           if((frcYR==NULL) || (frcUR==NULL) || (frcVR==NULL))
              {
                cout<<"Memory for R codes NOT allocated \n";
                exit(0);
              }
             allocateflag=1;
             return;
       }

public:
  _COLVIDFRCODE(){
     aNbitQ=bNbitQ=aNbitR=bNbitR=-1;
    codeflag=0;
    videoflag=0;
    allocateflag=0;
    GOP=-1;
    startFrame=endFrame=-1;
  }
  ~_COLVIDFRCODE(){}

    void setNumBitQ(int aNbit, int bNbit)
    {
         aNbitQ=aNbit;
         bNbitQ=bNbit;
    }

    void setNumBitR(int aNbit, int bNbit)
    {
         aNbitR=aNbit;
         bNbitR=bNbit;
    }

    void setStartFrameNo(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(allocateflag){
        cout<<"Memory allocated for frames between "<<startFrame<<" and "<<endFrame<<". NO CHANGE in the status. \n";
        return;
      }
      if((num<0) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between ZERO and "<<endFrame<<" \n";
       return;
      }
       startFrame=num;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return;
    }

    void setEndFrameNo(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(allocateflag){
        cout<<"Memory allocated for frames between "<<startFrame<<" and "<<endFrame<<". NO CHANGE in the status. \n";
        return;
      }
      if((num<startFrame) || (num>totalVideoFrameNo)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<totalVideoFrameNo<<" \n";
       return;
      }
       endFrame=num;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return;
    }

    int openVideoStream(char *fname,int w=176,int h=144)
    {
      videoflag=vid.open(fname,w,h);
      if(videoflag ==0) {
        cout<<"Unable to OPEN video stream \n";
        return 0;
      }
      width=w;
      height=h;
      totalVideoFrameNo=vid.getTotalFrameNo();
      startFrame=0;
      endFrame=totalVideoFrameNo-1;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return 1;
    }

    void setGOP(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(num<=0)
       {
        cout<<"GOP must be a POSITIVE NUMBER. GOP NOT SET.\n";
        return;
       }
      if(GOP>0) {
        cout<<"GOP already set. NO CHANGE in the status. \n";
        return;
      }
       GOP=num;
       totalQFrameNo=(totalCompressionFrameNo-1)/GOP+1;
       totalRFrameNo=totalCompressionFrameNo-totalQFrameNo;
       allocateFractalCodes();
       return;
    }

  int get_w() {return width;}
  int get_h() {return height;}
  int get_GOP() {return GOP;}
  int getTotalFrameNo() {return totalCompressionFrameNo;}
  int getTotalQFrameNo() {return totalQFrameNo;}
  int getTotalRFrameNo() {return totalRFrameNo;}
  int getStartFrame() {return startFrame;}
  int getEndFrame() {return endFrame;}


    void compress(int hrsize,int rsize,int err_lb,int l_gap=1,int rng_subst=0,float err_th=0)
{
  _GREY  imY,imU,imV;
   int i;
   int qIndex=-1, rIndex=-1;

    if(GOP<1) {
       cout<<"GOP NOT set. NOT encoded. \n";
       return;
    }


  HRNG_SIZE=hrsize;
  RNG_SIZE=rsize;
  ERR_LB=err_lb;
  ERR_TH=err_th;
  L_GAP=l_gap;


   for(i=startFrame;i<=endFrame;i++)
   {
       cout<<"Going to extract "<<i<<" th  frame \n"<<flush;
       imY=vid.getGreyFrameNo(i);
       imU=vid.getUFrameNo(i);
       imV=vid.getVFrameNo(i);
       cout<<i<<" th Y,U, V frames extracted \n"<<flush;
      if((i-startFrame)%GOP==0){
        qIndex++;
        if(qIndex>=totalQFrameNo){
         cout<<"WARNING! qIndex exceeding "<<totalQFrameNo<<"\n";
         exit(0);
        }
         frcYQ[qIndex].setNumBitAB(aNbitQ,bNbitQ);
         frcUQ[qIndex].setNumBitAB(aNbitQ,bNbitQ);
         frcVQ[qIndex].setNumBitAB(aNbitQ,bNbitQ);
        cout<<"Going to compress "<<i<<"th frame as Q frame \n";
        frcYQ[qIndex].compress(imY,HRNG_SIZE,RNG_SIZE,ERR_LB,L_GAP);
        frcUQ[qIndex].compress(imU,HRNG_SIZE,RNG_SIZE,ERR_LB,L_GAP);
        frcVQ[qIndex].compress(imV,HRNG_SIZE,RNG_SIZE,ERR_LB,L_GAP);
      }
     else{
       rIndex++;
        if(rIndex>=totalRFrameNo){
         cout<<"WARNING! rIndex exceeding "<<totalRFrameNo<<"\n";
         exit(0);
        }
         frcYR[rIndex].setNumBitAB(aNbitR,bNbitR);
         frcUR[rIndex].setNumBitAB(aNbitR,bNbitR);
         frcVR[rIndex].setNumBitAB(aNbitR,bNbitR);


           if((i-startFrame)%GOP==1){
         frcYR[rIndex].setFractalCode(frcYQ[qIndex]);
         frcUR[rIndex].setFractalCode(frcUQ[qIndex]);
         frcVR[rIndex].setFractalCode(frcVQ[qIndex]);
         }
         else
         {
         frcYR[rIndex].setFractalCode(frcYR[rIndex-1]);
         frcUR[rIndex].setFractalCode(frcUR[rIndex-1]);
         frcVR[rIndex].setFractalCode(frcVR[rIndex-1]);
         }
         // frcR[rIndex].setFractalCode(frcQ[qIndex]);

        cout<<"Going to compress "<<i<<"th frame as R frame \n";
         if(rng_subst)
         { 
          frcYR[rIndex].compress(imY,1,err_th); //with range block substitution
          frcUR[rIndex].compress(imU,1,err_th); //with range block substitution
          frcVR[rIndex].compress(imV,1,err_th); //with range block substitution
         }
          else
          {
          frcYR[rIndex].compress(imY,0,err_th); //without range block substitution
          frcUR[rIndex].compress(imU,0,err_th); //without range block substitution
          frcVR[rIndex].compress(imV,0,err_th); //without range block substitution
          }

     }
      // cout<<"Going to free video frame \n";
         /*
         imY.free();
         imU.free();
         imV.free();
         */
      // cout<<"Freed video frame \n";
   }

    codeflag=1;
    vid.close();
    videoflag=0;

 }

 _GREY decompressY(int num) // Decompresses num'th frame
  {
   _GREY im;
    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return(im);
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return im;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th Y Frame as Q Frame \n";
            im=frcYQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th Y Frame as R Frame \n";
           im=frcYR[(GOP-1)*nQ+nR-1].decompress();
      }
            return im;
  }
/*
 void decompressY(_GREY im,int num) // Decompresses num'th frame
  {
    int nQ,nR;

     if((im.get_h()!=height) || (im.get_w()!=width))
       {
         cout<<"Image dimension does not match ! \n";
         return;
       }

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return;
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return ;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th Y Frame as Q Frame \n";
            im=frcYQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th Y Frame as R Frame \n";
           im=frcYR[(GOP-1)*nQ+nR-1].decompress();
      }
  }
  */

 _GREY decompressU(int num) // Decompresses num'th frame
  {
   _GREY im;
    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return(im);
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return im;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th U Frame as Q Frame \n";
            im=frcUQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th U Frame as R Frame \n";
           im=frcUR[(GOP-1)*nQ+nR-1].decompress();
      }
            return im;
  }
/*
 void decompressU(_GREY im,int num) // Decompresses num'th frame
  {
    int nQ,nR;

     if((im.get_h()!=height/2) || (im.get_w()!=width/2))
       {
         cout<<"Image dimension does not match ! \n";
         return;
       }

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return;
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return ;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th Y Frame as Q Frame \n";
            im=frcUQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th Y Frame as R Frame \n";
           im=frcUR[(GOP-1)*nQ+nR-1].decompress();
      }
  }
 */

 _GREY decompressV(int num) // Decompresses num'th frame
  {
   _GREY im;
    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return(im);
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return im;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th V Frame as Q Frame \n";
            im=frcVQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th V Frame as R Frame \n";
           im=frcVR[(GOP-1)*nQ+nR-1].decompress();
      }
            return im;
  }
/*
 void decompressV(_GREY im,int num) // Decompresses num'th frame
  {
    int nQ,nR;

     if((im.get_h()!=height/2) || (im.get_w()!=width/2))
       {
         cout<<"Image dimension does not match ! \n";
         return;
       }

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return;
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return ;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th Y Frame as Q Frame \n";
            im=frcVQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th Y Frame as R Frame \n";
           im=frcVR[(GOP-1)*nQ+nR-1].decompress();
      }
  }
  */

 void decompressIntoFile(char *fname)
 {
    int i;
   _GREY imY,imU,imV;

   ofstream fout(fname);

   if(!fout){
    cout<<"Problem in creating file "<<fname<<"\n";
    return;
   }
   for(i=startFrame;i<=endFrame;i++)
    {
       imY=decompressY(i);
       imU=decompressU(i);
       imV=decompressV(i);
       imY.truncate_pic();
       imU.truncate_pic();
       imV.truncate_pic();
       fout<<imY<<imU<<imV;
        cout<<i<<"th frame written \n";
       /*
       imY.free();
       imU.free();
       imV.free();
       */
    }
    fout.close();
    return;
 }

void writeCode(char *fname)
   {
    int i;
    int qIndex=-1,rIndex=-1;
    ofstream fout(fname);

        if(fout.bad()) {
          cout<<"Problem in creating the file "<<fname<<" \n";
          return;
        }
   fout<<startFrame<<" "<<endFrame<<" "<<GOP<<"\n";
   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        frcYQ[qIndex].quantiseAB();
        frcUQ[qIndex].quantiseAB();
        frcVQ[qIndex].quantiseAB();
        fout<<frcYQ[qIndex];
        fout<<frcUQ[qIndex];
        fout<<frcVQ[qIndex];
      }
      else{
        rIndex++;
        frcYR[rIndex].quantiseAB();
        frcUR[rIndex].quantiseAB();
        frcVR[rIndex].quantiseAB();
        fout<<frcYR[rIndex];
        fout<<frcUR[rIndex];
        fout<<frcVR[rIndex];
      }


   fout.close();
   }

void readCode(char *fname)
   {
    int i;
    int qIndex=-1,rIndex=-1;
    ifstream fin(fname);

        if(fin.bad()) {
          cout<<"Problem in creating the file "<<fname<<" \n";
          return;
        }
   fin>>startFrame>>endFrame>>GOP;

       totalCompressionFrameNo=endFrame-startFrame+1;
       totalQFrameNo=(totalCompressionFrameNo-1)/GOP+1;
       totalRFrameNo=totalCompressionFrameNo-totalQFrameNo;
       allocateFractalCodes();

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        fin>>frcYQ[qIndex];
        fin>>frcUQ[qIndex];
        fin>>frcVQ[qIndex];
        cout<<"Read "<<i<<"th code as Q frame \n";
      }
      else{
        rIndex++;

           if((i-startFrame)%GOP==1){
         frcYR[rIndex].setFractalCode(frcYQ[qIndex]);
         frcUR[rIndex].setFractalCode(frcUQ[qIndex]);
         frcVR[rIndex].setFractalCode(frcVQ[qIndex]);
         }
         else
         {
         frcYR[rIndex].setFractalCode(frcYR[rIndex-1]);
         frcUR[rIndex].setFractalCode(frcUR[rIndex-1]);
         frcVR[rIndex].setFractalCode(frcVR[rIndex-1]);
         }
         // frcR[rIndex].setFractalCode(frcQ[qIndex]);

        fin>>frcYR[rIndex];
        fin>>frcUR[rIndex];
        fin>>frcVR[rIndex];
        cout<<"Read "<<i<<"th code as R frame \n";
      }
     width=frcYQ[0].get_w();
     height=frcYQ[0].get_h();

    codeflag=1;

   fin.close();
   }



   void printStat()
   {
    int i;
    int qIndex=-1,rIndex=-1;

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        frcYQ[qIndex].printStat();
        frcUQ[qIndex].printStat();
        frcVQ[qIndex].printStat();
      }
      else{
        rIndex++;
        frcYR[rIndex].printStat();
        frcUR[rIndex].printStat();
        frcVR[rIndex].printStat();
      }

   }

   float computeDataSize()// in bits 
   {
    int i;
    int qIndex=-1,rIndex=-1;
    float bppY,bppU,bppV;
    float tot_bppY=0;
    float tot_bppU=0;
    float tot_bppV=0;
    float dsizeY,dsizeU,dsizeV;

   if(codeflag==0){
     cout<<"WARNING! NOT CODED! RETURNING -1 \n";
     return -1;
   }

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        bppY=frcYQ[qIndex].compute_bpp();
        bppU=frcUQ[qIndex].compute_bpp();
        bppV=frcVQ[qIndex].compute_bpp();
        cout<<"For "<<i<<"th Q Frame bppY="<<bppY<<"\n";;
        cout<<"For "<<i<<"th Q Frame bppU="<<bppU<<"\n";;
        cout<<"For "<<i<<"th Q Frame bppV="<<bppV<<"\n";;
        tot_bppY+=bppY;
        tot_bppU+=bppU;
        tot_bppV+=bppV;
      }
      else{
        rIndex++;
        bppY=frcYR[rIndex].compute_bpp();
        bppU=frcUR[rIndex].compute_bpp();
        bppV=frcVR[rIndex].compute_bpp();
        cout<<"For "<<i<<"th R Frame bppY="<<bppY<<"\n";;
        cout<<"For "<<i<<"th R Frame bppU="<<bppU<<"\n";;
        cout<<"For "<<i<<"th R Frame bppV="<<bppV<<"\n";;
        tot_bppY+=bppY;
        tot_bppU+=bppU;
        tot_bppV+=bppV;
      }
       cout<<"Total bppY="<<tot_bppY<<"\n";
       cout<<"Total bppU="<<tot_bppU<<"\n";
       cout<<"Total bppV="<<tot_bppV<<"\n";

      dsizeY=tot_bppY*(float)width*(float)height;
      dsizeU=tot_bppU*(float)(width/2)*(float)(height/2);
      dsizeV=tot_bppV*(float)(width/2)*(float)(height/2);

      return(dsizeY+dsizeU+dsizeV);

   }

   float computeDataSize(int n)// in bits 
   {
    int i;
    int qIndex=-1,rIndex=-1;
    float bppY,bppU,bppV;
    float tot_bppY=0;
    float tot_bppU=0;
    float tot_bppV=0;
    float dsizeY,dsizeU,dsizeV;

   if(codeflag==0){
     cout<<"computeDataSize: WARNING! NOT CODED! RETURNING -1 \n";
     return -1;
   }
   if((startFrame+n-1)>endFrame)
    {
     cout<<"computeDataSize: WARNING! END FRAME EXCEEDED! RETURNING -1 \n";
     return -1;
    }

   for(i=startFrame;i<=(startFrame+n-1);i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        bppY=frcYQ[qIndex].compute_bpp();
        bppU=frcUQ[qIndex].compute_bpp();
        bppV=frcVQ[qIndex].compute_bpp();
        cout<<"For "<<i<<"th Q Frame bppY="<<bppY<<"\n";;
        cout<<"For "<<i<<"th Q Frame bppU="<<bppU<<"\n";;
        cout<<"For "<<i<<"th Q Frame bppV="<<bppV<<"\n";;
        tot_bppY+=bppY;
        tot_bppU+=bppU;
        tot_bppV+=bppV;
      }
      else{
        rIndex++;
        bppY=frcYR[rIndex].compute_bpp();
        bppU=frcUR[rIndex].compute_bpp();
        bppV=frcVR[rIndex].compute_bpp();
        cout<<"For "<<i<<"th R Frame bppY="<<bppY<<"\n";;
        cout<<"For "<<i<<"th R Frame bppU="<<bppU<<"\n";;
        cout<<"For "<<i<<"th R Frame bppV="<<bppV<<"\n";;
        tot_bppY+=bppY;
        tot_bppU+=bppU;
        tot_bppV+=bppV;
      }
       cout<<"Total bppY="<<tot_bppY<<"\n";
       cout<<"Total bppU="<<tot_bppU<<"\n";
       cout<<"Total bppV="<<tot_bppV<<"\n";

      dsizeY=tot_bppY*(float)width*(float)height;
      dsizeU=tot_bppU*(float)(width/2)*(float)(height/2);
      dsizeV=tot_bppV*(float)(width/2)*(float)(height/2);

      return(dsizeY+dsizeU+dsizeV);
   }

   void printStat(int num)
   {
    int i;

    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return;
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return ;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
            frcYQ[nQ].printStat();
            frcUQ[nQ].printStat();
            frcVQ[nQ].printStat();
         }
      else{
           frcYR[(GOP-1)*nQ+nR-1].printStat();
           frcUR[(GOP-1)*nQ+nR-1].printStat();
           frcVR[(GOP-1)*nQ+nR-1].printStat();
      }

   }

};
