/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2001). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
class _GRVIDFRCODE{
  int width,height;
  int HRNG_SIZE,RNG_SIZE,L_GAP;
  float ERR_LB,ERR_TH;
  int aNbitQ,bNbitQ;
  int aNbitR,bNbitR;
  _QDFSTDECODE *frcQ; // Pointer to array of I(Q) frames
  _RELFRCODE   *frcR; // Pointer to array of Relative Frames
  int codeflag; // 1 or 0
  int videoflag; // 1 or 0
  int allocateflag; // 1 or 0
  _QCIFVIDEO vid;
   int startFrame;
   int endFrame;
   int totalVideoFrameNo;
   int totalCompressionFrameNo;
   int GOP;
   int totalQFrameNo;
   int totalRFrameNo;

       void allocateFractalCodes()
       {
           if(GOP<1) {
             cout<<"GOP NOT DEFINED \n";
             return;
           }
           frcQ=(_QDFSTDECODE *)calloc(totalQFrameNo,sizeof(_QDFSTDECODE));
           if(frcQ==NULL)
              {
                cout<<"Memory for Q codes NOT allocated \n";
                exit(0);
              }
           frcR=(_RELFRCODE *)calloc(totalRFrameNo,sizeof(_RELFRCODE));
           if(frcR==NULL)
              {
                cout<<"Memory for R codes NOT allocated \n";
                exit(0);
              }
             allocateflag=1;
             return;
       }

public:
  _GRVIDFRCODE(){
     aNbitQ=bNbitQ=aNbitR=bNbitR=-1;
    codeflag=0;
    videoflag=0;
    allocateflag=0;
    GOP=-1;
    startFrame=endFrame=-1;
  }
  ~_GRVIDFRCODE(){}

    void setNumBitQ(int aNbit, int bNbit)
    {
         aNbitQ=aNbit;
         bNbitQ=bNbit;
    }

    void setNumBitR(int aNbit, int bNbit)
    {
         aNbitR=aNbit;
         bNbitR=bNbit;
    }

    void setStartFrameNo(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(allocateflag){
        cout<<"Memory allocated for frames between "<<startFrame<<" and "<<endFrame<<". NO CHANGE in the status. \n";
        return;
      }
      if((num<0) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between ZERO and "<<endFrame<<" \n";
       return;
      }
       startFrame=num;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return;
    }

    void setEndFrameNo(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(allocateflag){
        cout<<"Memory allocated for frames between "<<startFrame<<" and "<<endFrame<<". NO CHANGE in the status. \n";
        return;
      }
      if((num<startFrame) || (num>totalVideoFrameNo)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<totalVideoFrameNo<<" \n";
       return;
      }
       endFrame=num;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return;
    }

    int openVideoStream(char *fname,int w=176,int h=144)
    {
      videoflag=vid.open(fname,w,h);
      if(videoflag ==0) {
        cout<<"Unable to OPEN video stream \n";
        return 0;
      }
      width=w;
      height=h;
      totalVideoFrameNo=vid.getTotalFrameNo();
      startFrame=0;
      endFrame=totalVideoFrameNo-1;
      totalCompressionFrameNo=endFrame-startFrame+1;
      return 1;
    }

    void setGOP(int num)
    {
      if(videoflag==0){
         cout<<"NO VIDEO stream in the encoder \n";
         return;
      }
      if(num<=0)
       {
        cout<<"GOP must be a POSITIVE NUMBER. GOP NOT SET.\n";
        return;
       }
      if(GOP>0) {
        cout<<"GOP already set. NO CHANGE in the status. \n";
        return;
      }
       GOP=num;
       totalQFrameNo=(totalCompressionFrameNo-1)/GOP+1;
       totalRFrameNo=totalCompressionFrameNo-totalQFrameNo;
       allocateFractalCodes();
       return;
    }

  int get_w() {return width;}
  int get_h() {return height;}
  int get_GOP() {return GOP;}
  int getTotalFrameNo() {return totalCompressionFrameNo;}
  int getTotalQFrameNo() {return totalQFrameNo;}
  int getTotalRFrameNo() {return totalRFrameNo;}
  int getStartFrame() {return startFrame;}
  int getEndFrame() {return endFrame;}


    void compress(int hrsize,int rsize,int err_lb,int l_gap=1,int rng_subst=0,float err_th=0)
{
  _GREY  gim;
   int i;
   int qIndex=-1, rIndex=-1;

    if(GOP<1) {
       cout<<"GOP NOT set. NOT encoded. \n";
       return;
    }


  HRNG_SIZE=hrsize;
  RNG_SIZE=rsize;
  ERR_LB=err_lb;
  ERR_TH=err_th;
  L_GAP=l_gap;


   for(i=startFrame;i<=endFrame;i++)
   {
       cout<<"Going to extract "<<i<<" th  frame \n"<<flush;
       gim=vid.getGreyFrameNo(i);
       cout<<i<<" th grey frame extracted \n"<<flush;
      if((i-startFrame)%GOP==0){
        qIndex++;
        if(qIndex>=totalQFrameNo){
         cout<<"WARNING! qIndex exceeding "<<totalQFrameNo<<"\n";
         exit(0);
        }
         frcQ[qIndex].setNumBitAB(aNbitQ,bNbitQ);
        cout<<"Going to compress "<<i<<"th frame as Q frame \n";
        frcQ[qIndex].compress(gim,HRNG_SIZE,RNG_SIZE,ERR_LB,L_GAP);
      }
     else{
       rIndex++;
        if(rIndex>=totalRFrameNo){
         cout<<"WARNING! rIndex exceeding "<<totalRFrameNo<<"\n";
         exit(0);
        }
         frcR[rIndex].setNumBitAB(aNbitR,bNbitR);


           if((i-startFrame)%GOP==1){
         frcR[rIndex].setFractalCode(frcQ[qIndex]);
         }
         else
         {
         frcR[rIndex].setFractalCode(frcR[rIndex-1]);
         }
         // frcR[rIndex].setFractalCode(frcQ[qIndex]);

        cout<<"Going to compress "<<i<<"th frame as R frame \n";
         if(rng_subst)
          frcR[rIndex].compress(gim,1,err_th); //with range block substitution
          else
          frcR[rIndex].compress(gim,0,err_th); //without range block substitution

     }
      // cout<<"Going to free video frame \n";
         // gim.free();
      // cout<<"Freed video frame \n";
   }

    codeflag=1;
    vid.close();
    videoflag=0;

 }

 _GREY decompress(int num) // Decompresses num'th frame
  {
   _GREY im;
    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return(im);
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return im;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
           cout<<"Decompressing "<<num<<"th Frame as Q Frame \n";
            im=frcQ[nQ].fastDecompress();
         }
      else{
           cout<<"Decompressing "<<num<<"th Frame as R Frame \n";
           im=frcR[(GOP-1)*nQ+nR-1].decompress();
      }
            return im;
  }

void writeCode(char *fname)
   {
    int i;
    int qIndex=-1,rIndex=-1;
    ofstream fout(fname);

        if(fout.bad()) {
          cout<<"Problem in creating the file "<<fname<<" \n";
          return;
        }
   fout<<startFrame<<" "<<endFrame<<" "<<GOP<<"\n";
   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        frcQ[qIndex].quantiseAB();
        fout<<frcQ[qIndex];
      }
      else{
        rIndex++;
        frcR[rIndex].quantiseAB();
        fout<<frcR[rIndex];
      }


   fout.close();
   }

void readCode(char *fname)
   {
    int i;
    int qIndex=-1,rIndex=-1;
    ifstream fin(fname);

        if(fin.bad()) {
          cout<<"Problem in creating the file "<<fname<<" \n";
          return;
        }
   fin>>startFrame>>endFrame>>GOP;

       totalCompressionFrameNo=endFrame-startFrame+1;
       totalQFrameNo=(totalCompressionFrameNo-1)/GOP+1;
       totalRFrameNo=totalCompressionFrameNo-totalQFrameNo;
       allocateFractalCodes();

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        fin>>frcQ[qIndex];
        cout<<"Read "<<i<<"th code as Q frame \n";
      }
      else{
        rIndex++;

           if((i-startFrame)%GOP==1){
         frcR[rIndex].setFractalCode(frcQ[qIndex]);
         }
         else
         {
         frcR[rIndex].setFractalCode(frcR[rIndex-1]);
         }
         // frcR[rIndex].setFractalCode(frcQ[qIndex]);

        fin>>frcR[rIndex];
        cout<<"Read "<<i<<"th code as R frame \n";
      }
     width=frcQ[0].get_w();
     height=frcQ[0].get_h();

    codeflag=1;

   fin.close();
   }



   void printStat()
   {
    int i;
    int qIndex=-1,rIndex=-1;

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        frcQ[qIndex].printStat();
      }
      else{
        rIndex++;
        frcR[rIndex].printStat();
      }

   }

   float computeDataSize()// in bits 
   {
    int i;
    int qIndex=-1,rIndex=-1;
    float bpp;
    float tot_bpp=0;

   if(codeflag==0){
     cout<<"WARNING! NOT CODED! RETURNING -1 \n";
     return -1;
   }

   for(i=startFrame;i<=endFrame;i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        bpp=frcQ[qIndex].compute_bpp();
        cout<<"For "<<i<<"th Q Frame bpp="<<bpp<<"\n";;
        tot_bpp+=bpp;
      }
      else{
        rIndex++;
        bpp=frcR[rIndex].compute_bpp();
        cout<<"For "<<i<<"th R Frame bpp="<<bpp<<"\n";;
        tot_bpp+=bpp;
      }
       cout<<"Total bpp="<<tot_bpp<<"\n";

      return(tot_bpp*(float)width*(float)height);

   }

   float computeDataSize(int n)// in bits 
   {
    int i;
    int qIndex=-1,rIndex=-1;
    float bpp;
    float tot_bpp=0;

   if(codeflag==0){
     cout<<"computeDataSize: WARNING! NOT CODED! RETURNING -1 \n";
     return -1;
   }
   if((startFrame+n-1)>endFrame)
    {
     cout<<"computeDataSize: WARNING! END FRAME EXCEEDED! RETURNING -1 \n";
     return -1;
    }

   for(i=startFrame;i<=(startFrame+n-1);i++)
      if((i-startFrame)%GOP==0){
        qIndex++;
        bpp=frcQ[qIndex].compute_bpp();
        cout<<"For "<<i<<"th Q Frame bpp="<<bpp<<"\n";;
        tot_bpp+=bpp;
      }
      else{
        rIndex++;
        bpp=frcR[rIndex].compute_bpp();
        cout<<"For "<<i<<"th R Frame bpp="<<bpp<<"\n";;
        tot_bpp+=bpp;
      }
       cout<<"Total bpp="<<tot_bpp<<"\n";

      return(tot_bpp*(float)width*(float)height);

   }

   void printStat(int num)
   {
    int i;

    int nQ,nR;

     if(codeflag==0){
     cout<<"WARNING! NOT encoded. Returning NULL image \n";
     return;
     }
      if((num<startFrame) || (num>endFrame)){
         cout<<"Invalid Parameter Range. Frame must be between "<<startFrame<<" and "<<endFrame<<" \n";
     cout<<"WARNING! Returning NULL image \n";
       return ;
      }

      nQ=(num-startFrame)/GOP;
      nR=(num-startFrame)%GOP;
      if(nR==0)
         {
            frcQ[nQ].printStat();
         }
      else{
           frcR[(GOP-1)*nQ+nR-1].printStat();
      }

   }

};
