/*************************************************************************
 These codes are developed by Jayanta Mukherjee, Dept. of CSE, IIT Kharagpur
 (between 1997-2001). The codes may be reused for any research and academic
 purpose. An acknowledgement of their uses would be appreciated.
 *************************************************************************/
class _QDFRCODE{
   int codeflag;
  int width,height;
  int RNG_SIZE,HRNG_SIZE,DOM_SIZE;  //Low range size, high range size
  int L_GAP;  // Lattice of separation for domain pool
  float ERR_LB;
  int **Rx,**Ry,**Tx;
  float **a,**b,**err;
  int **Rqd;
  float aMin,aMax,bMin,bMax;
  int aNbit, bNbit; // -1 means floating point representation
  int dxNbit, dyNbit; // Bits for Domain addresses 
  int txNbit; // Bit for the transformation
  float aQnt,bQnt;  // 0 means floating point representation
  struct _qdbyte{
         unsigned char bit0:1;
         unsigned char bit1:1;
         unsigned char bit2:1;
         unsigned char bit3:1;
         unsigned char bit4:1;
         unsigned char bit5:1;
         unsigned char bit6:1;
         unsigned char bit7:1;
         /*
         unsigned int bit0:1;
         unsigned int bit1:1;
         unsigned int bit2:1;
         unsigned int bit3:1;
         unsigned int bit4:1;
         unsigned int bit5:1;
         unsigned int bit6:1;
         unsigned int bit7:1;
         */
         } qdinfo;
  struct _txbyte{
          unsigned short int tx0:3;
          unsigned short int tx1:3;
          unsigned short int tx2:3;
          unsigned short int tx3:3;
          unsigned short int tx4:3;
          /*
          unsigned short int tx5:3;
          unsigned int tx6:3;
          unsigned int tx7:3;
           */
         } txinfo;

      int allocate_coeff_arrays()
       {
              a=allocate_matrix_float(height/RNG_SIZE,width/RNG_SIZE,sizeof(float));
              if(a==NULL)
                 {
                  cout<<"COEFF a's not allocated \n";
                  return -1;
                 }
              b=allocate_matrix_float(height/RNG_SIZE,width/RNG_SIZE,sizeof(float));
              if(b==NULL)
                 {
                  cout<<"COEFF b's not allocated \n";
                  return -1;
                 }

              err=allocate_matrix_float(height/RNG_SIZE,width/RNG_SIZE,sizeof(float));
              if(err==NULL)
                 {
                  cout<<"Error array not allocated \n";
                  return -1;
                 }

              Rx=allocate_matrix_int(height/RNG_SIZE,width/RNG_SIZE,sizeof(int));
              if(Rx==NULL)
                 {
                  cout<<"COEFF Rx's not allocated \n";
                  return -1;
                 }

              Ry=allocate_matrix_int(height/RNG_SIZE,width/RNG_SIZE,sizeof(int));
              if(Ry==NULL)
                 {
                  cout<<"COEFF Ry's not allocated \n";
                  return -1;
                 }

              Tx=allocate_matrix_int(height/RNG_SIZE,width/RNG_SIZE,sizeof(int));
              if(Tx==NULL)
                 {
                  cout<<"COEFF Tx's not allocated \n";
                  return -1;
                 }

              Rqd=allocate_matrix_int(height/RNG_SIZE,width/RNG_SIZE,sizeof(int));
              if(Rqd==NULL)
                 {
                  cout<<"Quad info array not allocated \n";
                  return -1;
                 }
              return 1;
       }

    void Initialize_arrays()
       {
         int i,j;

              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
               {
                a[i][j]=b[i][j]=0;
                err[i][j]=99999.0;
                Rqd[i][j]=0;
                Rx[i][j]=Ry[i][j]=Tx[i][j]=-1;
               }

       }

   int check_quadrant(int x,int y,int factor)
        {
           int i,j;
           int flag=1;
         
           if(factor==1) return 1;

           for(i=x;i<x+factor;i++)
           for(j=y;j<y+factor;j++)
            {
             if(err[i][j]>ERR_LB)
              {
              flag=0;
              }
             if(!flag) break;
            }
              if(flag)
                err[x][y]=99999.0; // For enforcing a code 
            return flag;
        }
  void adjustFactor(int x,int y,int factor)
       {
         int i,j;
          // if(factor>1)
          // cout<<"adjusting at "<<x<<","<<y<<" with "<<factor<<"\n";
           for(i=x;i<x+factor;i++)
           for(j=y;j<y+factor;j++)
               Rqd[i][j]=-1;
            Rqd[x][y]=factor;
       }

  void compute_factor_in_Rqd()
       {
         int i,j,k,found,lx,ly;
         int count_zero;

              for(i=0;i<height/RNG_SIZE;i++)
              {
                found=0;
                count_zero=0;
              for(j=0;j<width/RNG_SIZE;j++)
               {
                 if(Rqd[i][j]<0) {
                         if(found)
                         if(j==(width/RNG_SIZE-1))
                           adjustFactor(lx,ly,count_zero);
                  }    //continue;
                 else
                 if(Rqd[i][j]==0) {
                         count_zero++;
                         if(j==(width/RNG_SIZE-1))
                           adjustFactor(lx,ly,count_zero);
                          }
                 else{
                     if(found) {
                       adjustFactor(lx,ly,count_zero);
                      }
                     lx=i; ly=j;
                     found=1;
                       count_zero=1;
                         if(j==(width/RNG_SIZE-1))
                           adjustFactor(lx,ly,count_zero);
                 }
               }
              }
       }

  void save_code(int x,int y,int dx,int dy,int tx,float ta,float tb,float terror,int factor)
         {
           int i,j;

           for(i=x;i<x+factor;i++)
           for(j=y;j<y+factor;j++)
               {
               Rx[i][j]=Ry[i][j]=Tx[i][j]=-1;
               Rqd[i][j]=0;
               a[i][j]=b[i][j]=0;
               err[i][j]=terror;
               }
            Rqd[x][y]=factor;
            Rx[x][y]=dx;
            Ry[x][y]=dy;
            Tx[x][y]=tx;
            a[x][y]=ta;
            b[x][y]=tb;
            err[x][y]=terror;

            return;
         }

 void getMinMaxAB()
         {
           int i,j;
          aMin=bMin=9999999.0;
          aMax=bMax=-9999999.0;
          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
           {
             if(Rqd[i][j]>0)
             {
               if(a[i][j]<aMin) aMin=a[i][j];
               if(a[i][j]>aMax) aMax=a[i][j];
               if(b[i][j]<bMin) bMin=b[i][j];
               if(b[i][j]>bMax) bMax=b[i][j];
             }
           }
          return;
         }

       int getNbit(int val)
       {
         int i=1,count=0;;
         if(val==1) return 1;
         while(i<val) {i=i<<1; count++;}
         return count;
       }



 public:

    _QDFRCODE(){width=height=RNG_SIZE=HRNG_SIZE=DOM_SIZE=codeflag=0;
              a=b=err=NULL;
              Rx=Ry=Tx=Rqd=NULL;

               aNbit=bNbit=-1;
               dxNbit=dyNbit=8;
               txNbit=3;
               aQnt=bQnt=0;
            } 

    ~_QDFRCODE(){} 
     void free()
       {
         if(a!=NULL) free_matrix_float(a,height/RNG_SIZE);
         if(b!=NULL) free_matrix_float(b,height/RNG_SIZE);
         if(err!=NULL) free_matrix_float(err,height/RNG_SIZE);
         if(Rx!=NULL) free_matrix_int(Rx,height/RNG_SIZE);
         if(Ry!=NULL) free_matrix_int(Ry,height/RNG_SIZE);
         if(Tx!=NULL) free_matrix_int(Tx,height/RNG_SIZE);
         if(Rqd!=NULL) free_matrix_int(Rqd,height/RNG_SIZE);
       }

 void setNumBitAB(int aBit,int bBit)
        {
           aNbit=aBit;
           bNbit=bBit;
           cout<<"Number of bits for a="<<aNbit<<" For b="<<bNbit<<"\n";
           return;
        }
 int **getQdHead()
       {
         return Rqd;
       }
 int get_w() {return width;}
 int get_h() {return height;}
 int get_rsize() {return RNG_SIZE;}
 float **get_a() {return a;}
 float **get_b() {return b;}
 int **get_Tx() {return Tx;}
 int **get_Rx() {return Rx;}
 int **get_Ry() {return Ry;}
 
 void computeNbitDxDy()
      {
           dxNbit= getNbit(width/L_GAP);
           dyNbit= getNbit(height/L_GAP);
           cout<<"Number of bits for Dx="<<dxNbit<<" For Dy="<<dyNbit<<"\n";
      }

  void quantiseAB()
        {
         long int aNumber=1,bNumber=1;
         int i,j;
         char ch;

         if((aNbit== -1) || (bNbit==-1))
          {
           cout<<"No. of bits for a and b not set \n";
           return;
          }
          aNumber=aNumber<<aNbit;
          bNumber=bNumber<<bNbit;
          cout<<"aNumber ="<<aNumber<<"\n";
           //cin>>ch;

          aQnt=(aMax-aMin)/(float)aNumber;
          bQnt=(bMax-bMin)/(float)bNumber;
          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
          {
           if(Rqd[i][j]>0)
            {
             a[i][j]= (float)((int)((a[i][j]-aMin)/aQnt));
             b[i][j]= (float)((int)((b[i][j]-bMin)/bQnt));
            }
          }
          return;
        }

   void dequantiseAB()
        {
         int i,j;
         if((aNbit== -1) || (bNbit==-1))
          {
           cout<<"No. of bits for a and b not set \n";
           return;
          }
          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
          {
           if(Rqd[i][j]>0)
            {
             a[i][j]= a[i][j]*aQnt+aMin;
             b[i][j]= b[i][j]*bQnt+bMin;
            }
          }
          return;
        }

  void printStat()
       {
           int i,j;
           int count[256];
           int total_count=0;
           int hfactor=HRNG_SIZE/RNG_SIZE;
           float bpp;

             //cout<<"Going to print the Statistics with highest factor "<<hfactor<<"\n";
            for(i=1;i<=hfactor;i++) count[i]=0;
             //cout<<"Count Initialised \n"<<flush;

              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
               if(Rqd[i][j]>0) count[Rqd[i][j]]++;
             //cout<<"Going to print the counts \n"<<flush;

        for(i=1;i<=hfactor;i++)
         if(count[i]>0)
          {
           total_count+=count[i];
           cout<<"No. of Blocks with size "<<RNG_SIZE*i<<" = "<<count[i]<<"\n"<<flush;
           }
           cout<<"Total No. of Blocks "<<total_count<<"\n"<<flush;
         computeNbitDxDy();
        if((aNbit!=-1) && (bNbit!=-1))
         bpp=(float)total_count * (float) (aNbit+bNbit+dxNbit+dyNbit+txNbit) / ((float)height*(float)width);
         else
         bpp=(float)total_count * (16+16+dxNbit+dyNbit+txNbit) / ((float)height*(float)width);
         cout<<"BPP for QDFC :"<<bpp<<"\n"<<flush;

       return;
       }




   void compress(_GREY im,int hrsize,int lrsize,float error,int l_gap=1)
       {
           if(codeflag==1) {
             cout<<"WARNING! Valid codes overwritten. \n";
           }

          _MASK domain,shDomain, range,temp_shDomain;
          float tcorr;
          int i,j,k,l,m;
          float sdDom,meanDom;
          float sdRng,meanRng;
          int cur_rsize,factor;
          int domain_found;


              width=im.get_w();
              height=im.get_h();
              RNG_SIZE=lrsize;
              HRNG_SIZE=hrsize;
              ERR_LB=error;
              codeflag=0;
              cur_rsize=RNG_SIZE;
              L_GAP=l_gap;

              allocate_coeff_arrays();
              Initialize_arrays();
           while(cur_rsize <=HRNG_SIZE)
              {
               cout<<"For range size "<<cur_rsize<<":\n";
               factor=cur_rsize/RNG_SIZE;
               DOM_SIZE=cur_rsize*2;


              domain.set(DOM_SIZE,DOM_SIZE);
              shDomain.set(cur_rsize,cur_rsize);
              range.set(cur_rsize,cur_rsize);
              temp_shDomain.set(cur_rsize,cur_rsize);

              for(i=0;i<height/cur_rsize;i++)
              for(j=0;j<width/cur_rsize;j++)
              {
                if(check_quadrant(i*factor,j*factor,factor))
                  {

                     domain_found=0;

                    im.get_Rect(range,i*cur_rsize,j*cur_rsize,cur_rsize);
                    sdRng=range.N_compute(SD_MODE);
                    meanRng=range.N_compute(AVG_MODE);

                    for(l=0;l<height-DOM_SIZE;l+=L_GAP)
                    for(m=0;m<width-DOM_SIZE;m+=L_GAP)
                    {
                       im.get_Rect(domain,l,m,DOM_SIZE);
                       domain.shrink(shDomain,2.0);
                       sdDom=shDomain.N_compute(SD_MODE);
                       meanDom=shDomain.N_compute(AVG_MODE);
                         

                       if (err[i*factor][j*factor]<=ERR_LB) break;
                       if((l!=i*cur_rsize) && (m!=j*cur_rsize) && (sdDom!=0))
                       {
          
                       if(sdRng!=0)
                       {
                        shDomain.copy(temp_shDomain);
                       for(k=0;k<7;k++){
                       tcorr=(range.avg_product(temp_shDomain)-meanRng*meanDom)/(sdRng*sdDom);
                       error=sqr(sdRng)*(1-tcorr*tcorr);
                       if( (error<err[i*factor][j*factor])|| (error<ERR_LB)) {
                               float tb,ta;
                               ta=tcorr*sdRng/sdDom;
                               tb=meanRng-ta*meanDom;
                               if(abs(ta)<1.0)
                               {
                               domain_found=1;
                              save_code(i*factor,j*factor,l,m,k,ta,tb,error,factor);
                               }
                               }
                             switch(k){
                             case 0:
                                       shDomain.rotate_90(temp_shDomain);
                                       break;
                             case 1:
                                       shDomain.rotate_180(temp_shDomain);
                                       break;
                             case 2:
                                       shDomain.rotate_270(temp_shDomain);
                                       break;
                             case 3:
                                       shDomain.reflect_y(temp_shDomain);
                                       break;
                             case 4:
                                       shDomain.rotate_180_and_reflect_y(temp_shDomain);
                                       break;
                             case 5:
                                       shDomain.transpose(temp_shDomain);
                                       break;
                             case 6:
                                       shDomain.reverse_transpose(temp_shDomain);
                                       break;
                             case 7:
                                       cout<<"Transform no. out of range \n";
                                       exit(0);
                             }
                         }
                       }
                       else
                       {
                             domain_found=1;
                    save_code(i*factor,j*factor,l,m,0,0,meanDom,0,factor);
                       }
                    }
                 } 

                 if(domain_found)
                 {
                 cout<<"For range at "<<i*factor<<","<<j*factor<<" Domain at"<<Rx[i*factor][j*factor]<<","<<Ry[i*factor][j*factor]<<"With Error"<<err[i*factor][j*factor]<<"\n";
                 }
               }//if check_quad 
              }

                 domain.free();
                 shDomain.free();
                 range.free();
                 temp_shDomain.free();
                 cout<<"Freed the temporary arrays \n";

               cur_rsize*=2;
             } //while

  cout<<"Number of blocks as follows:\n";
           printStat();
           getMinMaxAB();
          cout<<"aMin= "<<aMin<<" aMax= "<<aMax<<"\n";
          cout<<"bMin= "<<bMin<<" bMax= "<<bMax<<"\n";
           quantiseAB();


                 codeflag=1;
                 return;
             }

      void writeCode(char *fname)
          {
            int i,j,k;
            ofstream fout(fname);

            fout<<width<<" "<<height<<" "<<RNG_SIZE<<" "<<HRNG_SIZE<<" "<<L_GAP<<"\n";
            fout<<aMin<<" "<<aQnt<<"\n";
            fout<<bMin<<" "<<bQnt<<"\n";
            fout<<aNbit<<" "<<bNbit<<"\n";


            for(i=0;i<height/RNG_SIZE;i++)
              {
              for(j=0;j<width/RNG_SIZE;j++)
                {
                  fout<<Rqd[i][j]<<" ";
                }
                 fout<<"\n";
              }


              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
                 if(Rqd[i][j]>0)
                  fout<<Rx[i][j]<<" "<<Ry[i][j]<<" "<<a[i][j]<<" "<<Tx[i][j]<<" "<<b[i][j]<<'\n';
            fout.close();
          }

      void writeBinCode(char *fname)
          {
            int i,j,k;
            int tmptxno;
            unsigned short int usint;
            unsigned char uchar;
            ofstream fout(fname);

           fout.write(&width,sizeof(int));
           fout.write(&height,sizeof(int));
           fout.write(&RNG_SIZE,sizeof(int));
           fout.write(&HRNG_SIZE,sizeof(int));
           fout.write(&L_GAP,sizeof(int));
           fout.write(&aMin,sizeof(float));
           fout.write(&bMin,sizeof(float));
           fout.write(&aQnt,sizeof(float));
           fout.write(&bQnt,sizeof(float));
           fout.write(&aNbit,sizeof(int));
           fout.write(&bNbit,sizeof(int));


            for(i=0;i<height/RNG_SIZE;i++)
              {
              for(j=0;j<width/RNG_SIZE;j+=8)
                {
                for(k=0;k<8;k++)
                  switch(k){
                  case 0: qdinfo.bit0=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 1: qdinfo.bit1=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 2: qdinfo.bit2=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 3: qdinfo.bit3=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 4: qdinfo.bit4=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 5: qdinfo.bit5=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 6: qdinfo.bit6=((Rqd[i][j+k]>0)?1:0);
                          break;
                  case 7: qdinfo.bit7=((Rqd[i][j+k]>0)?1:0);
                          break;
                  default:
                          cout<<"Value of k exceeds 7 \n";
                          exit(0);
                  }
           fout.write(&qdinfo,sizeof(struct _qdbyte));
                }
              }


          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
           if(Rqd[i][j]>0)
           {
           uchar=(unsigned char) (Rx[i][j]/L_GAP);
           fout.write(&uchar,sizeof(unsigned char));

           uchar=(unsigned char) (Ry[i][j]/L_GAP);
           fout.write(&uchar,sizeof(unsigned char));
           }

          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
           if(Rqd[i][j]>0)
           {
           if(a[i][j]==256) a[i][j]=255;
           uchar=(unsigned char) a[i][j];
           fout.write(&uchar,sizeof(unsigned char));

           if(b[i][j]==256) b[i][j]=255;
           uchar=(unsigned char) b[i][j];
           fout.write(&uchar,sizeof(unsigned char));
           }

           tmptxno=-1;

          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
           if(Rqd[i][j]>0)
           {
             /*
           uchar=(unsigned char) Tx[i][j];
           fout.write(&uchar,sizeof(unsigned char));
              */
            tmptxno++;
              if(tmptxno==5){
                fout.write(&txinfo,sizeof(struct _txbyte));
                tmptxno=0;
              }
              switch(tmptxno){
              case 0: txinfo.tx0=Tx[i][j];
                      break;
              case 1: txinfo.tx1=Tx[i][j];
                      break;
              case 2: txinfo.tx2=Tx[i][j];
                      break;
              case 3: txinfo.tx3=Tx[i][j];
                      break;
              case 4: txinfo.tx4=Tx[i][j];
                      break;
             /*
              case 5: txinfo.tx5=Tx[i][j];
                      break;
              case 6: txinfo.tx6=Tx[i][j];
                      break;
              case 7: txinfo.tx7=Tx[i][j];
                      break;
              */
              default: 
                      cout<<"Transformation number out of range \n";
                      exit(0);
              }
           }
            if((tmptxno>=0) && (tmptxno<5))
                fout.write(&txinfo,sizeof(struct _txbyte));
            fout.close();
          }

       void readCode(char *fname)
          {
            int i,j;
            ifstream fin(fname);

                cout<<"Going to read code from file "<<fname<<"\n"<<flush;
            fin>>width>>height>>RNG_SIZE>>HRNG_SIZE>>L_GAP;
            fin>>aMin>>aQnt;
            fin>>bMin>>bQnt;
            fin>>aNbit>>bNbit;


              allocate_coeff_arrays();
              cout<<"Memory for coeffcients allocated \n"<<flush;

              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
                {
                   fin>>Rqd[i][j];
                }


              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
               {
                  if(Rqd[i][j]>0)
                  fin>>Rx[i][j]>>Ry[i][j]>>a[i][j]>>Tx[i][j]>>b[i][j];
                   else
                    {
                      Rx[i][j]=Ry[i][j]=-1;
                      a[i][j]=b[i][j]=0;
                      Tx[i][j]=-1;
                    }
               }
             codeflag=1;
            fin.close();
          }

      void readBinCode(char *fname)
          {
            int i,j,k;
            unsigned char uchar;
            unsigned short int usint;
            int tmptxno;
            ifstream fin(fname);

           fin.read(&width,sizeof(int));
           fin.read(&height,sizeof(int));
           fin.read(&RNG_SIZE,sizeof(int));
           fin.read(&HRNG_SIZE,sizeof(int));
           fin.read(&L_GAP,sizeof(int));
           fin.read(&aMin,sizeof(float));
           fin.read(&bMin,sizeof(float));
           fin.read(&aQnt,sizeof(float));
           fin.read(&bQnt,sizeof(float));
           fin.read(&aNbit,sizeof(int));
           fin.read(&bNbit,sizeof(int));

              allocate_coeff_arrays();
              cout<<"Memory for coeffcients allocated \n"<<flush;

            for(i=0;i<height/RNG_SIZE;i++)
              {
              for(j=0;j<width/RNG_SIZE;j+=8)
                {
                 fin.read(&qdinfo,sizeof(struct _qdbyte));
                for(k=0;k<8;k++)
                  {
                  switch(k){
                  case 0: Rqd[i][j+k]=qdinfo.bit0;
                          break;
                  case 1: Rqd[i][j+k]=qdinfo.bit1;
                          break;
                  case 2: Rqd[i][j+k]=qdinfo.bit2;
                          break;
                  case 3: Rqd[i][j+k]=qdinfo.bit3;
                          break;
                  case 4: Rqd[i][j+k]=qdinfo.bit4;
                          break;
                  case 5: Rqd[i][j+k]=qdinfo.bit5;
                          break;
                  case 6: Rqd[i][j+k]=qdinfo.bit6;
                          break;
                  case 7: Rqd[i][j+k]=qdinfo.bit7;
                          break;
                  default:
                          cout<<"Value of k exceeds 7 \n";
                          exit(0);
                   }
                   //cout<<" "<<Rqd[i][j+k];
                  }
                  //cout<<"\n";
                }
              }

           cout<<"Going to compute factors ..\n";
        compute_factor_in_Rqd();
           cout<<"Factors computed ..\n";

           /*
            for(i=0;i<height/RNG_SIZE;i++)
              {
              for(j=0;j<width/RNG_SIZE;j++)
                    cout<<" "<<Rqd[i][j];
              cout<<"\n";
            }
           */


              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
                 if(Rqd[i][j]>0)
                 {
                   fin.read(&uchar,sizeof(unsigned char));
                     Rx[i][j]= L_GAP*(unsigned int)uchar;

                     fin.read(&uchar,sizeof(unsigned char));
                     Ry[i][j]= L_GAP*(unsigned int)uchar;
                 }
                   else
                    {
                      Rx[i][j]=Ry[i][j]=-1;
                    }

              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
                 if(Rqd[i][j]>0)
                 {
                 fin.read(&uchar,sizeof(unsigned char));
                 a[i][j]=(float) uchar;

                 fin.read(&uchar,sizeof(unsigned char));
                 b[i][j]=(float) uchar;
                 }
                   else
                    {
                      a[i][j]=b[i][j]=0;
                    }
              tmptxno=5;

              for(i=0;i<height/RNG_SIZE;i++)
              for(j=0;j<width/RNG_SIZE;j++)
                 if(Rqd[i][j]>0)
                 {
                  /*
                 fin.read(&uchar,sizeof(unsigned char));
                 Tx[i][j]= (unsigned int)uchar;
                  */
                      if(tmptxno==5)
                      {
                      fin.read(&txinfo,sizeof(struct _txbyte));
                      tmptxno=0;
                      }
                      switch(tmptxno)
                      {
                       case 0: Tx[i][j]=txinfo.tx0;
                               break;
                       case 1: Tx[i][j]=txinfo.tx1;
                               break;
                       case 2: Tx[i][j]=txinfo.tx2;
                               break;
                       case 3: Tx[i][j]=txinfo.tx3;
                               break;
                       case 4: Tx[i][j]=txinfo.tx4;
                               break;
                       /*
                       case 5: Tx[i][j]=txinfo.tx5;
                               break;
                       case 6: Tx[i][j]=txinfo.tx6;
                               break;
                       case 7: Tx[i][j]=txinfo.tx7;
                               break;
                        */
                       default:
                                cout<<"Transformation Index out of range \n";
                                exit(0);
                      }
                      tmptxno++;

                 }
                  else
                      Tx[i][j]=-1;
             codeflag=1;
            fin.close();
          }

     _GREY  decompress(int ITER_NO=30)
         {

          _GREY im;
          _MASK domain,shDomain,temp_shDomain;
          int i,j;
          int iter;
          int cur_rsize,factor;

          if(codeflag==0){
             cout<<"WARNING! Not containing any codes! Returning NULL image \n";
           return(im);
          }
          dequantiseAB();

          im.set(width,height,100,DFLOAT);
          cout<<"Image Initialised \n";

          for(iter=0;iter<ITER_NO;iter++)
          {
            cur_rsize=RNG_SIZE;
            while(cur_rsize<=HRNG_SIZE)
            {
              DOM_SIZE=2*cur_rsize;
              factor=cur_rsize/RNG_SIZE;

                domain.set(DOM_SIZE,DOM_SIZE);
                shDomain.set(cur_rsize,cur_rsize);
                temp_shDomain.set(cur_rsize,cur_rsize);


          for(i=0;i<height/RNG_SIZE;i++)
          for(j=0;j<width/RNG_SIZE;j++)
          {
           if(Rqd[i][j]==factor)
            {
           // if(factor>1) cout<<"Domain size for Range Block at ("<<i<<","<<j<<") :"<<DOM_SIZE<<"\n";
           im.get_Rect(domain,Rx[i][j],Ry[i][j],DOM_SIZE);
           domain.shrink(shDomain,2.0);
           im.apply_affine_to(i*RNG_SIZE,j*RNG_SIZE,shDomain, a[i][j],b[i][j],Tx[i][j]);
            }
           }
           cur_rsize*=2;

          domain.free();
          shDomain.free();
          temp_shDomain.free();
           }
          cout<<"iteration no. "<<iter<<'\n';
          }
            return(im);
         }
      };
