

#define REVDATE   "Version 3.00  Rev: 3/30/93"


/* include files */
#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stddef.h>
#include <stdlib.h>


/* include the appropriate string header file */ 
#include <strings.h>

#include <sys/types.h>    /* only include <sys/types.h> once */



/* indicies into conv24MB */
#define CONV24_8BIT  0
#define CONV24_24BIT 1
#define CONV24_SEP1  2
#define CONV24_LOCK  3
#define CONV24_SEP2  4
#define CONV24_FAST  5
#define CONV24_SLOW  6
#define CONV24_BEST  7
#define CONV24_MAX   8

/* values 'picType' can take */
#define PIC8  CONV24_8BIT
#define PIC24 CONV24_24BIT



/* constants for setting radio buttons in dirW */
#define F_COLORS    0
#define F_FORMAT    1

#define F_FULLCOLOR 0
#define F_GREYSCALE 1
#define F_BWDITHER  2
#define F_REDUCED   3

#define F_GIF       0
#define F_PM        1
#define F_PBMRAW    2
#define F_PBMASCII  3
#define F_XBM       4
#define F_SUNRAS    5
#define F_BMP       6
#define F_PS        7
#define F_IRIS      8

typedef unsigned char byte;

/* info structure filled in by the LoadXXX() image reading routines */
typedef struct { byte *pic;                  /* image data */
		 int   w, h;                 /* size */
		 int   type;                 /* PIC8 or PIC24 */

		 byte  r[256],g[256],b[256];
		                             /* colormap, if PIC8 */

		 int   frmType;              /* def. Format type to save in */
		 int   colType;              /* def. Color type to save in */
		 char  fullInfo[128];        /* Format: field in info box */
		 char  shrtInfo[128];        /* short format info */
		 char *comment;              /* comment text */
		 
		 int   numpages;             /* # of page files, if >1 */
		 char  pagebname[64];        /* basename of page files */
	       } PICINFO;


/* MACROS */

/* RANGE forces a to be in the range b..c (inclusive) */
#define RANGE(a,b,c) { if (a < b) a = b;  if (a > c) a = c; }

/* PTINRECT returns '1' if x,y is in rect (inclusive) */
#define PTINRECT(x,y,rx,ry,rw,rh) \
           ((x)>=(rx) && (y)>=(ry) && (x)<=(rx)+(rw) && (y)<=(ry)+(rh))

/* MONO returns total intensity of r,g,b components */
#define MONO(rd,gn,bl) (((rd)*11 + (gn)*16 + (bl)*5) >> 5)  /*.33R+ .5G+ .17B*/

/* ISPIPE returns true if the passed in character is considered the
   start of a 'load-from-pipe' or 'save-to-pipe' string */
#define ISPIPE(c) ((c)=='!' || (c)=='|')


#define  RAW 0
#define   PBM 1
#define   PPM 2
#define   PGM 3

#define TRUNCSTR "File appears to be truncated."

 int garbage;
 long numgot, filesize;
 char bname[100];
int src_mode=RAW, dst_mode=PGM;
PICINFO *pinfo;

/*******************************************/
int LoadPBM(fname, pinfo)
     char    *fname;
     PICINFO *pinfo;
/*******************************************/
{
  /* returns '1' on success */

  FILE  *fp;
  int    c, c1;
  int    maxv, rv;

  printf("Into LoadPBM() \n");
  garbage = maxv = 0;
  /* bname = BaseName(fname); */
  pinfo->pic     = (byte *) NULL;
  pinfo->comment = (char *) NULL;

  /* open the file */
  fp=fopen(fname,"r");
  if (!fp) return (pbmError(bname, "can't open file"));
   else printf("File %s opened \n",fname);

  /* compute file length */
  fseek(fp, 0L, 2);
  filesize = ftell(fp);
  fseek(fp, 0L, 0);

  printf("File length =%ld \n",filesize);


  /* read the first two bytes of the file to determine which format
     this file is.  "P1" = ascii bitmap, "P2" = ascii greymap,
     "P3" = ascii pixmap, "P4" = raw bitmap, "P5" = raw greymap,
     "P6" = raw pixmap */

  c = getc(fp);  c1 = getc(fp);
  if (c!='P' || c1<'1' || c1>'6') return(pbmError(bname, "unknown format"));

  /* read in header information */
  pinfo->w = Jgetint(fp, pinfo);  pinfo->h = Jgetint(fp, pinfo);

  /* if we're not reading a bitmap, read the 'max value' */
  if ( !(c1=='1' || c1=='4')) {
    maxv = Jgetint(fp, pinfo);
    if (maxv < 1) garbage=1;    /* to avoid 'div by zero' probs */
  }


  if (garbage) {
    fclose(fp);
    if (pinfo->comment) free(pinfo->comment);
    pinfo->comment = (char *) NULL;
    return (pbmError(bname, "Garbage characters in header."));
  }


  if (c1=='1' || c1=='2' || c1=='3') pinfo->frmType = F_PBMASCII;
                                else pinfo->frmType = F_PBMRAW;

  /* note:  pic, type, r,g,b, frmInfo, shortFrm, and colorType fields of
     picinfo struct are filled in in the format-specific loaders */

  /* call the appropriate subroutine to handle format-specific stuff */
  if      (c1=='1' || c1=='4') rv = loadpbm(fp, pinfo, c1=='4' ? 1 : 0);
  else if (c1=='2' || c1=='5') rv = loadpgm(fp, pinfo, c1=='5' ? 1 : 0, maxv);
  else if (c1=='3' || c1=='6') rv = loadppm(fp, pinfo, c1=='6' ? 1 : 0, maxv);

  fclose(fp);

  if (!rv) {
    if (pinfo->pic) free(pinfo->comment);
    if (pinfo->comment) free(pinfo->comment);
    pinfo->pic     = (byte *) NULL;
    pinfo->comment = (char *) NULL;
  }

  return rv;
}  


/*******************************************/
int LoadRAW(fname, pinfo)
     char    *fname;
     PICINFO *pinfo;
/*******************************************/
{
  /* returns '1' on success */

  FILE  *fp;
  int    c, c1;
  int    maxv, rv;
  byte *pix, *pic8;
  int i,j,w,h,numgot;
  int raw=1;
  int color=1;

  garbage = maxv = 0;

  pinfo->pic     = (byte *) NULL;
  pinfo->comment = (char *) NULL;


  /* open the file */
  fp=fopen(fname,"r");
  if (!fp) return (pbmError(bname, "can't open file"));

  /* compute file length */
  fseek(fp, 0L, 2);
  filesize = ftell(fp);
  fseek(fp, 0L, 0);


  /* read in header information */
  printf("Width and height? \n");
  scanf("%d%d",&pinfo->w,&pinfo->h);


  w = pinfo->w;  h = pinfo->h;

   color=filesize/(w*h);
  pic8 = (byte *) calloc(w*h*color,1);
  if (!pic8) return(pbmError(bname, "couldn't malloc 'pic8'"));


  pinfo->pic  = pic8;
  if(color==1)
  pinfo->type = PIC8;
  else
  pinfo->type = PIC24;
  sprintf(pinfo->fullInfo, "RAW, %s format.  (%ld bytes)", 
	  (raw) ? "raw" : "ascii", filesize);
  sprintf(pinfo->shrtInfo, "%dx%d RAW", pinfo->w, pinfo->h);
  if(color==1)
  pinfo->colType = F_GREYSCALE;
  else
  pinfo->colType = F_FULLCOLOR;

  /* fill in a greyscale colormap where maxv maps to 255 */
  for (i=0; i<=255; i++)
    pinfo->r[i] = pinfo->g[i] = pinfo->b[i] = (byte) i;



  numgot = fread(pic8, 1, w*h*color, fp);   /* read raw data */

  if (numgot != w*h*color) pbmError(bname, TRUNCSTR);   /* warning only */

  if (garbage) {
    return (pbmError(bname, "Garbage characters in image data."));
  }

  fclose(fp);

  return 1;
}  



/*******************************************/
 int loadpbm(fp, pinfo, raw)
     FILE    *fp;
     PICINFO *pinfo;
     int      raw;
{
  byte *pic8;
  byte *pix;
  int   i,j,bit,w,h;

   printf("Reading PBM file \n");
  w = pinfo->w;  h = pinfo->h;
  pic8 = (byte *) calloc(w * h, 1);
  if (!pic8) return pbmError(bname, "couldn't malloc 'pic8'");

  pinfo->pic  = pic8;
  pinfo->type = PIC8;
  sprintf(pinfo->fullInfo, "PBM, %s format.  (%ld bytes)", 
	  (raw) ? "raw" : "ascii", filesize);
  sprintf(pinfo->shrtInfo, "%dx%d PBM.", w, h);
  pinfo->colType = F_BWDITHER;


  /* B/W bitmaps have a two entry colormap */
  pinfo->r[0] = pinfo->g[0] = pinfo->b[0] = 255;   /* entry #0 = white */
  pinfo->r[1] = pinfo->g[1] = pinfo->b[1] = 0;     /* entry #1 = black */


  if (!raw) {
    numgot = 0;
    for (i=0, pix=pic8; i<h; i++) {
      if ((i&0x3f)==0) WaitCursor();
      for (j=0; j<w; j++, pix++) *pix = Jgetbit(fp, pinfo);
    }

    if (numgot != w*h) pbmError(bname, TRUNCSTR);
    if (garbage) {
      return(pbmError(bname, "Garbage characters in image data."));
    }
  }


  else {   /* read raw bits */
    int trunc = 0, k = 0;

    for (i=0, pix=pic8; i<h; i++) {
      if ((i&15)==0) WaitCursor();
      for (j=0,bit=0; j<w; j++, pix++, bit++) {

	bit &= 7;
	if (!bit) {
	  k = getc(fp);
	  if (k==EOF) { trunc=1; k=0; }
	}

	*pix = (k&0x80) ? 1 : 0;
	k = k << 1;
      }
    }

    if (trunc) pbmError(bname, TRUNCSTR);
  }

  return 1;
}


/*******************************************/
 int loadpgm(fp, pinfo, raw, maxv)
     FILE    *fp;
     PICINFO *pinfo;
     int      raw, maxv;
{
  byte *pix, *pic8;
  int   i,j,bitshift,w,h;


   printf("Reading PGM file \n");
  w = pinfo->w;  h = pinfo->h;
  pic8 = (byte *) calloc(w*h,1);
  if (!pic8) return(pbmError(bname, "couldn't malloc 'pic8'"));


  pinfo->pic  = pic8;
  pinfo->type = PIC8;
  sprintf(pinfo->fullInfo, "PGM, %s format.  (%ld bytes)", 
	  (raw) ? "raw" : "ascii", filesize);
  sprintf(pinfo->shrtInfo, "%dx%d PGM.", pinfo->w, pinfo->h);
  pinfo->colType = F_GREYSCALE;


  /* if maxv>255, keep dropping bits until it's reasonable */
  bitshift = 0;
  while (maxv>255) { maxv = maxv>>1;  bitshift++; }

  /* fill in a greyscale colormap where maxv maps to 255 */
  for (i=0; i<=maxv; i++)
    pinfo->r[i] = pinfo->g[i] = pinfo->b[i] = (i*255)/maxv;


  if (!raw) {
    numgot = 0;
    for (i=0, pix=pic8; i<h; i++) {
      if ((i&0x3f)==0) WaitCursor();
      for (j=0; j<w; j++, pix++)
	*pix = (Jgetint(fp, pinfo) >> bitshift);
    }
  }

  else numgot = fread(pic8, 1, w*h, fp);   /* read raw data */

  if (numgot != w*h) pbmError(bname, TRUNCSTR);   /* warning only */

  if (garbage) {
    return (pbmError(bname, "Garbage characters in image data."));
  }

  return 1;
}


/*******************************************/
 int loadppm(fp, pinfo, raw, maxv)
     FILE    *fp;
     PICINFO *pinfo;
     int      raw, maxv;
{
  byte *pix, *pic24, scale[256], *pic8;
  int   i,j,bitshift, w, h;

   printf("Reading PPM file \n");
  w = pinfo->w;  h = pinfo->h;

  /* allocate 24-bit image */
  pic24 = (byte *) calloc(w*h*3,1);
  if (!pic24) FatalError("couldn't malloc 'pic24'");

  pinfo->pic  = pic24;
  pinfo->type = PIC24;
  sprintf(pinfo->fullInfo, "PPM, %s format.  (%ld bytes)", 
	  (raw) ? "raw" : "ascii", filesize);
  sprintf(pinfo->shrtInfo, "%dx%d PPM.", w, h);
  pinfo->colType = F_FULLCOLOR;
  printf("%s\n%s\n",pinfo->fullInfo,pinfo->shrtInfo);


  /* if maxv>255, keep dropping bits until it's reasonable */
  bitshift = 0;
  while (maxv>255) { maxv = maxv>>1;  bitshift++; }


  if (!raw) {
    numgot = 0;
    for (i=0, pix=pic24; i<h; i++) {
      if ((i&0x3f)==0) WaitCursor();

      for (j=0; j<w*3; j++, pix++)
	*pix = (Jgetint(fp, pinfo) >> bitshift);
    }
  }

  else numgot = fread(pic24, 1, w*h*3, fp);    /* read raw data */


  if (numgot != w*h*3) pbmError(bname, TRUNCSTR);

  if (garbage)
    return(pbmError(bname, "Garbage characters in image data."));


  /* have to scale all RGB values up (Conv24to8 expects RGB values to
     range from 0-255 */

  if (maxv<255) { 
    printf("maxv<255\n");
    for (i=0; i<=maxv; i++) scale[i] = (i * 255) / maxv;

    for (i=0, pix=pic24; i<h; i++) {
      if ((i&0x3f)==0) WaitCursor();
      for (j=0; j<w*3; j++, pix++) *pix = scale[*pix];
    }
  }

  return 1;
}



/*******************************************/
 int Jgetint(fp, pinfo)
     FILE *fp;
     PICINFO *pinfo;
{
  int c, i, firstchar;

  /* note:  if it sees a '#' character, all characters from there to end of
     line are appended to the comment string */

  /* skip forward to start of next number */
  c = getc(fp);
  while (1) {
    /* eat comments */
    if (c=='#') {   /* if we're at a comment, read to end of line */
      char cmt[256], *sp, *tmpptr;

      sp = cmt;  firstchar = 1;
      while (1) {
	c=getc(fp);
	if (firstchar && c == ' ') firstchar = 0;  /* lop off 1 sp after # */
	else {
	  if (c == '\n' || c == EOF) break;
	  if ((sp-cmt)<250) *sp++ = c;
	}
      }
      *sp++ = '\n';
      *sp   = '\0';

      if (strlen(cmt) > 0) {    /* add to pinfo->comment */
	if (!pinfo->comment) {
	  pinfo->comment = (char *) malloc(strlen(cmt)+1);
	  if (!pinfo->comment) FatalError("malloc failure in xvpbm.c Jgetint");
	  pinfo->comment[0] = '\0';
	}
	else {
	  tmpptr = (char *) realloc(pinfo->comment, 
		      strlen(pinfo->comment) + strlen(cmt) + 1); 
	  if (!tmpptr) FatalError("realloc failure in xvpbm.c Jgetint");
	  pinfo->comment = tmpptr;
	}
	strcat(pinfo->comment, cmt);
      }
    }

    if (c==EOF) return 0;
    if (c>='0' && c<='9') break;   /* we've found what we were looking for */

    /* see if we are getting garbage (non-whitespace) */
    if (c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',') garbage=1;

    c = getc(fp);
  }


  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fp);
    if (c==EOF) return i;
    if (c<'0' || c>'9') break;
  }

  numgot++;
  return i;
}



/*******************************************/
 int Jgetbit(fp, pinfo)
     FILE *fp;
     PICINFO *pinfo;
{
  int c;

  /* skip forward to start of next number */
  c = getc(fp);
  while (1) {
    /* eat comments */
    if (c=='#') {   /* if we're at a comment, read to end of line */
      char cmt[256], *sp, *tmpptr;

      sp = cmt;
      while (1) {
	c=getc(fp);
	if (c == '\n' || c == EOF) break;

	if ((sp-cmt)<250) *sp++ = c;
      }
      *sp++ = '\n';
      *sp = '\0';

      if (strlen(cmt) > 0) {    /* add to pinfo->comment */
	if (!pinfo->comment) {
	  pinfo->comment = (char *) malloc(strlen(cmt)+1);
	  if (!pinfo->comment) FatalError("malloc failure in xvpbm.c Jgetint");
	  pinfo->comment[0] = '\0';
	}
	else {
	  tmpptr = (char *) realloc(pinfo->comment, 
		      strlen(pinfo->comment) + strlen(cmt) + 1); 
	  if (!tmpptr) FatalError("realloc failure in xvpbm.c Jgetint");
	  pinfo->comment = tmpptr;
	}
	strcat(pinfo->comment, cmt);
      }
    }
    if (c==EOF) return 0;
    if (c=='0' || c=='1') break;   /* we've found what we were looking for */

    /* see if we are getting garbage (non-whitespace) */
    if (c!=' ' && c!='\t' && c!='\r' && c!='\n' && c!=',') garbage=1;

    c = getc(fp);
  }


  numgot++;
  return(c-'0');
}


/*******************************************/
/*
 int pbmError(fname, st)
     char *fname, *st;
{
  SetISTR(ISTR_WARNING,"%s:  %s", fname, st);
  return 0;
}
*/
 int pbmError(fname, st)
     char *fname, *st;
{
  printf("Error: %s \n",st);
  return 0;
}

 int FatalError( st)
     char  *st;
{
  printf("Error: %s \n",st);
  exit(0);
}




/*******************************************/
int WritePBM(fp,pic,ptype,w,h,rmap,gmap,bmap,numcols,colorstyle,raw,comment)
     FILE *fp;
     byte *pic;
     int   ptype, w,h;
     byte *rmap, *gmap, *bmap;
     int   numcols, colorstyle, raw;
     char *comment;
{
  /* writes a PBM/PGM/PPM file to the already open stream
     if (raw), writes as RAW bytes, otherwise writes as ASCII 
     'colorstyle' single-handedly determines the type of file written
     if colorstyle==0, (Full Color) a PPM file is written
     if colorstyle==1, (Greyscale)  a PGM file is written
     if colorstyle==2, (B/W stipple) a PBM file is written */

  int   magic;
  byte *pix;
  int   i,j,len;

  /* calc the appropriate magic number for this file type */
  magic = 0;
  if      (colorstyle==0) magic = 3;
  else if (colorstyle==1) magic = 2;
  else if (colorstyle==2) magic = 1;

  if (raw && magic) magic+=3;


  /* write the header info */
  fprintf(fp,"P%d\n",magic);
  fprintf(fp,"# CREATOR: XV %s\n", REVDATE);

  if (comment) {      /* write comment lines */
    char *sp;

    sp = comment;
    while (*sp) {
      fprintf(fp, "# ");
      while (*sp && *sp != '\n') fputc(*sp++, fp);
      if (*sp == '\n') sp++;
      fputc('\n', fp);
    }
  }


  fprintf(fp,"%d %d\n",w,h);
  if (colorstyle!=2) fprintf(fp,"255\n");

  if (ferror(fp)) return -1;
  if(ptype==PIC8)
      printf("PIC8 Format \n");
  else 
      printf("PIC24 Format \n");

  /* write the image data */

  if (colorstyle==0) {                  /* 24bit RGB, 3 bytes per pixel */
    for (i=0, pix=pic, len=0; i<h; i++) {
      if ((i&63)==0) WaitCursor();
      for (j=0; j<w; j++) {
	if (raw) {
	  if (ptype==PIC8) {
	    putc(rmap[*pix],fp);  putc(gmap[*pix],fp);  putc(bmap[*pix],fp);
	  }
	  else {  /* PIC24 */
	    putc(pix[0],fp);  putc(pix[1],fp);  putc(pix[2],fp);
	  }
	}
	else {
	  if (ptype==PIC8) 
	    fprintf(fp,"%3d %3d %3d ",rmap[*pix], gmap[*pix], bmap[*pix]);
	  else
	    fprintf(fp,"%3d %3d %3d ",pix[0], pix[1], pix[2]);

	  len+=12;
	  if (len>58) { fprintf(fp,"\n");  len=0; }
	}
	
	pix += (ptype==PIC24) ? 3 : 1;
      }
    }
  }


  else if (colorstyle==1) {             /* 8-bit greyscale */
    byte rgb[256];
    if (ptype==PIC8)
      for (i=0; i<numcols; i++) rgb[i] = MONO(rmap[i],gmap[i],bmap[i]);
    for (i=0, pix=pic, len=0; i<w*h; i++) {
      if ((i&0x7fff)==0) WaitCursor();

      if (raw) putc((ptype==PIC8) ? rgb[*pix] : MONO(pix[0],pix[1],pix[2]),fp);

      else {
	if (ptype==PIC8) fprintf(fp,"%3d ",rgb[*pix]);
	            else fprintf(fp,"%3d ",MONO(pix[0],pix[1],pix[2]));
	len += 4;
	if (len>66) { fprintf(fp,"\n");  len=0; }
      }

      pix += (ptype==PIC24) ? 3 : 1;
    }
  }

  else if (colorstyle==2) {             /* 1-bit B/W stipple */
    int bit,k,flipbw;
    char *str0, *str1;

    /* shouldn't happen */
    if (ptype == PIC24) FatalError("PIC24 and B/W Stipple in WritePBM()\n");

    /* if '0' is black, set flipbw */
    flipbw = (MONO(rmap[0],gmap[0],bmap[0]) < MONO(rmap[1],gmap[1],bmap[1]));

    str0 = (flipbw) ? "1 " : "0 ";
    str1 = (flipbw) ? "0 " : "1 ";

    for (i=0, pix=pic, len=0; i<h; i++) {
      if ((i&15)==0) WaitCursor();
      for (j=0, bit=0, k=0; j<w; j++, pix++) {
	if (raw) {
	  k = (k << 1) | *pix;
	  bit++;
	  if (bit==8) {
	    if (flipbw) k = ~k;
	    fputc(k,fp);
	    bit = k = 0;
	  }
	}
	else {
	  if (*pix) fprintf(fp,str1);
	       else fprintf(fp,str0);
	  len+=2;
	  if (len>68) { fprintf(fp,"\n"); len=0; }
	}
      } /* j */
      if (raw && bit) {
	k = k << (8-bit);
	if (flipbw) k = ~k;
	fputc(k,fp);
      }
    }
  }

  if (ferror(fp)) return -1;

  return 0;
}

WriteRAW(fp,pic,ptype,w,h,rmap,gmap,bmap,colorstyle)
FILE *fp;
byte *pic,*rmap,*gmap,*bmap;
int ptype,w,h;
{

  /* writes a RAW data   to the already open stream
     if (raw), writes as RAW bytes, otherwise writes as ASCII */ 
  

  int   magic;
  byte *pix;
  int   i,j,len;


  printf("width= %d height= %d\n",w,h);

  if (ferror(fp)) return -1;

  /* write the image data */

  if (colorstyle==0) {                  /* 24bit RGB, 3 bytes per pixel */
    for (i=0, pix=pic, len=0; i<h; i++) {
      if ((i&63)==0) WaitCursor();
      for (j=0; j<w; j++) {
	  if (ptype==PIC8) {
	    putc(rmap[*pix],fp);  putc(gmap[*pix],fp);  putc(bmap[*pix],fp);
	  }
          else {
	    putc(pix[0],fp);   putc(pix[1],fp);   putc(pix[2],fp);
          }
        pix += (ptype==PIC24) ? 3 : 1; 
	}

      }
    }

  else if (colorstyle==1) {             /* 8-bit greyscale */
    byte rgb[256];
    if (ptype==PIC8)
      for (i=0; i<256; i++) rgb[i] = MONO(rmap[i],gmap[i],bmap[i]);

    for (i=0, pix=pic, len=0; i<w*h; i++) {
      if ((i&0x7fff)==0) WaitCursor();

       putc((ptype==PIC8) ? rgb[*pix] : MONO(pix[0],pix[1],pix[2]),fp);

      pix += (ptype==PIC24) ? 3 : 1;
    }
  }

  else if (colorstyle==2) {             /* 1-bit B/W stipple */
    int bit,k,flipbw;
    char *str0, *str1;

    /* shouldn't happen */
    if (ptype == PIC24) FatalError("PIC24 and B/W Stipple in WritePBM()\n");

    /* if '0' is black, set flipbw */
    flipbw = (MONO(rmap[0],gmap[0],bmap[0]) < MONO(rmap[1],gmap[1],bmap[1]));

    str0 = (flipbw) ? "1 " : "0 ";
    str1 = (flipbw) ? "0 " : "1 ";

    for (i=0, pix=pic, len=0; i<h; i++) {
      if ((i&15)==0) WaitCursor();
      for (j=0, bit=0, k=0; j<w; j++, pix++) {
	  k = (k << 1) | *pix;
	  bit++;
	  if (bit==8) {
	    if (flipbw) k = ~k;
	    fputc(k,fp);
	    bit = k = 0;
	  }
      } /* j */
	k = k << (8-bit);
	if (flipbw) k = ~k;
	fputc(k,fp);
    }
  }

  if (ferror(fp)) return -1;

  return 0;
}


WaitCursor()
{
return;
}
	  
	  

main(argc,argv)
int argc;
char *argv[];
{
 FILE *sfp,*dfp;
 char *src_fname, *dst_fname;
 int raw,colorstyle,numcols;
 char comment[100];


 comment[0]='\0';
 if(argc<2) {
  printf("argc=%d \n",argc);
  printf("Usage: conv_pbm [-rp/-rg/-rb/-pr...] <src_file> <dest_file> \n");
  exit(0);
  }
 if(argc==3){
  src_mode=RAW;
  dst_mode=PGM;
  src_fname=argv[1];
  dst_fname=argv[2];
  /*
  if((sfp=fopen(argv[1],"rb"))==NULL){
   printf("Error in opening input file %s \n",argv[1]);
   exit(0);
   }
  */
  if((dfp=fopen(argv[2],"wb"))==NULL){
   printf("Error in opening output file %s \n",argv[2]);
   exit(0);
   }
 }
 else {
  printf("argc=%d \n",argc);
  if(argv[1][0]=='-'){
  switch(argv[1][1]){
  case 'r': src_mode=RAW;
            break;
  case 'g': src_mode=PGM;
            break;
  case 'b': src_mode=PBM;
            break;
  case 'p': src_mode=PPM;
            break;
  default:   src_mode=0;
  }
  switch(argv[1][2]){
  case 'r': dst_mode=RAW;
            break;
  case 'g': dst_mode=PGM;
            break;
  case 'b': dst_mode=PBM;
            break;
  case 'p': dst_mode=PPM;
            break;
  default:   dst_mode=0;
  }
  if(src_mode==dst_mode) {
  printf("No conversion needed. \n");
  exit(0);
  }
  src_fname=argv[2];
  dst_fname=argv[3];
  /*
  if((sfp=fopen(argv[2],"rb"))==NULL){
   printf("Error in opening input file %s \n",argv[1]);
   exit(0);
   }
   */
  if((dfp=fopen(argv[3],"wb"))==NULL){
   printf("Error in opening output file %s \n",argv[2]);
   exit(0);
   }
  }
  else {
  printf("Usage: conv_pbm [-r/-p] <src_file> <dest_file> \n");
  exit(0);
  }
 }
  printf("Src_mode= %d dst_mode=%d \n",src_mode,dst_mode);
  printf("Source =%s Dest. =%s \n",src_fname,dst_fname);
  pinfo=(PICINFO *) malloc(sizeof(PICINFO));

switch(src_mode){
case RAW:
         LoadRAW(src_fname,pinfo);
         break;
case PGM:
         colorstyle=1;
         LoadPBM(src_fname,pinfo);
         break;
case PPM:
         colorstyle=0;
         LoadPBM(src_fname,pinfo);
         break;
case PBM:
         colorstyle=2;
         LoadPBM(src_fname,pinfo);
         break;
}
printf("File loaded \n");
raw=1;
switch(dst_mode){
case RAW:
        WriteRAW(dfp,pinfo->pic,pinfo->type,pinfo->w,pinfo->h,pinfo->r,pinfo->g,pinfo->b,colorstyle);
        break;
case PGM:
         colorstyle=1;
         WritePBM(dfp,pinfo->pic,pinfo->type,pinfo->w,pinfo->h,pinfo->r,pinfo->g,pinfo->b,256,colorstyle,raw,comment);
         break;
case PPM:
         colorstyle=0;
         WritePBM(dfp,pinfo->pic,pinfo->type,pinfo->w,pinfo->h,pinfo->r,pinfo->g,pinfo->b,256,colorstyle,raw,comment);
         break;
case PBM:
         colorstyle=2;
         WritePBM(dfp,pinfo->pic,pinfo->type,pinfo->w,pinfo->h,pinfo->r,pinfo->g,pinfo->b,2,colorstyle,raw,comment);
         break;
}
 fclose(dfp);

}


